from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from infrahub.core.constants import RelationshipDirection
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.migrations.shared import MigrationInput, MigrationResult
from infrahub.core.query import Query, QueryType
from infrahub.log import get_logger

from ..shared import ArbitraryMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class CleanUpDuplicatedUuidVertices(Query):
    """
    Find vertexes that include the given label and have the same UUID and same set of labels
    For each of these duplicate vertex groups, keep one and mark all the others to be deleted by the PerformHardDeletesQuery
      - Group all of the edges touching a vertex in this vertex group by branch, edge_type, peer_element_id, and direction
        - For each edge group, we will link one edge to the vertex we are keeping for this vertex group and mark all of the others to be deleted
        - we will set/create one active edge from the vertex to keep to the peer of this group, setting its from time to the earliest active
            from time in this group
        - if ALL edges in this edge group are deleted, then we will set the to time of the active edge to the latest deleted time and
            set/create a deleted edge with a from time of the latest deleted time
    """

    name = "clean_up_duplicated_uuid_vertices"
    type = QueryType.WRITE
    insert_return = False
    insert_limit = False

    def __init__(
        self,
        vertex_label: str,
        outbound_edge_types: list[DatabaseEdgeType],
        inbound_edge_types: list[DatabaseEdgeType],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.vertex_label = vertex_label
        self.outbound_edge_types = outbound_edge_types
        self.inbound_edge_types = inbound_edge_types

    def _get_or_create_active_edge_subquery(
        self,
        edge_type: DatabaseEdgeType,
        direction: Literal[RelationshipDirection.INBOUND, RelationshipDirection.OUTBOUND],
    ) -> str:
        if direction is RelationshipDirection.INBOUND:
            l_arrow = "<"
            r_arrow = ""
        else:
            l_arrow = ""
            r_arrow = ">"

        query = """
    CALL (vertex_to_keep, edge_type, branch, peer, earliest_active_time, latest_deleted_time, all_edges_deleted, edge_to_copy) {
        // ------------
        // get or create the active %(edge_type)s edge
        // ------------
        WITH edge_type
        WHERE edge_type = "%(edge_type)s"
        MERGE (vertex_to_keep)%(l_arrow)s-[active_edge:%(edge_type)s {branch: branch, status: "active"}]-%(r_arrow)s(peer)
        WITH active_edge
        LIMIT 1
        SET active_edge.to_delete = NULL
        SET active_edge.from = earliest_active_time
        SET active_edge.to = CASE
            WHEN all_edges_deleted = TRUE THEN latest_deleted_time
            ELSE NULL
        END
        SET active_edge.branch_level = edge_to_copy.branch_level
        SET active_edge.hierarchy = edge_to_copy.hierarchy
    }
        """ % {
            "edge_type": edge_type.value,
            "l_arrow": l_arrow,
            "r_arrow": r_arrow,
        }
        return query

    def _add_deleted_edge_subquery(
        self,
        edge_type: DatabaseEdgeType,
        direction: Literal[RelationshipDirection.INBOUND, RelationshipDirection.OUTBOUND],
    ) -> str:
        if direction is RelationshipDirection.INBOUND:
            l_arrow = "<"
            r_arrow = ""
        else:
            l_arrow = ""
            r_arrow = ">"
        subquery = """
    CALL (vertex_to_keep, edge_type, branch, peer, latest_deleted_time, edge_to_copy) {
        // ------------
        // create the deleted %(edge_type)s edge
        // ------------
        WITH edge_type
        WHERE edge_type = "%(edge_type)s"
        MERGE (vertex_to_keep)%(l_arrow)s-[deleted_edge:%(edge_type)s {branch: branch, status: "deleted"}]-%(r_arrow)s(peer)
        WITH deleted_edge
        LIMIT 1
        SET deleted_edge.to_delete = NULL
        SET deleted_edge.from = latest_deleted_time
        SET deleted_edge.to = NULL
        SET deleted_edge.branch_level = edge_to_copy.branch_level
        SET deleted_edge.hierarchy = edge_to_copy.hierarchy
    }
        """ % {"edge_type": edge_type.value, "l_arrow": l_arrow, "r_arrow": r_arrow}
        return subquery

    def _build_directed_edges_subquery(
        self,
        db: InfrahubDatabase,
        direction: Literal[RelationshipDirection.INBOUND, RelationshipDirection.OUTBOUND],
        edge_types: list[DatabaseEdgeType],
    ) -> str:
        if direction is RelationshipDirection.INBOUND:
            l_arrow = "<"
            r_arrow = ""
        else:
            l_arrow = ""
            r_arrow = ">"
        active_subqueries: list[str] = []
        delete_subqueries: list[str] = []
        for edge_type in edge_types:
            active_subqueries.append(
                self._get_or_create_active_edge_subquery(
                    edge_type=edge_type,
                    direction=direction,
                )
            )
            delete_subqueries.append(self._add_deleted_edge_subquery(edge_type=edge_type, direction=direction))
        active_edge_subqueries = "\n".join(active_subqueries)
        deleted_edge_subqueries = "\n".join(delete_subqueries)

        edges_query = """
//------------
// Get every %(direction)s branch, edge_type, peer element_id combinations touching vertices with this uuid/labels combination
//------------
CALL (n_uuid, vertex_element_ids, element_id_to_keep) {
    CALL (n_uuid, vertex_element_ids) {
        MATCH (n:%(vertex_label)s {uuid: n_uuid})
        WHERE %(id_func_name)s(n) IN vertex_element_ids
        MATCH (n)%(l_arrow)s-[e]-%(r_arrow)s(peer)
        WITH DISTINCT e.branch AS branch, type(e) AS edge_type, %(id_func_name)s(peer) AS peer_element_id
        RETURN branch, edge_type, peer_element_id
    }
    //------------
    // Are all of the edges with these with this branch/edge_type/peer_element_id combination deleted?
    //------------
    CALL (n_uuid, vertex_element_ids, branch, edge_type, peer_element_id) {
        // nodes with this edge_type/branch/peer combo
        MATCH (node_with_edge:%(vertex_label)s {uuid: n_uuid})%(l_arrow)s-[e {branch: branch}]-%(r_arrow)s(peer)
        WHERE %(id_func_name)s(node_with_edge) IN vertex_element_ids
        AND type(e) = edge_type
        AND %(id_func_name)s(peer) = peer_element_id
        // count of nodes with this edge_type/branch/peer combo
        WITH DISTINCT n_uuid, branch, edge_type, peer_element_id, %(id_func_name)s(node_with_edge) AS node_with_edge_element_id
        WITH n_uuid, branch, edge_type, peer_element_id, collect(node_with_edge_element_id) AS node_with_edge_element_ids
        // nodes with this edge_type/branch/peer combo where the edge is DELETED
        OPTIONAL MATCH (node_with_deleted_edge:%(vertex_label)s {uuid: n_uuid})%(l_arrow)s-[e {branch: branch}]-%(r_arrow)s(peer)
        WHERE %(id_func_name)s(node_with_deleted_edge) IN node_with_edge_element_ids
        AND type(e) = edge_type
        AND %(id_func_name)s(peer) = peer_element_id
        AND (e.status = "deleted" OR e.to IS NOT NULL)
        // count of nodes with this DELETED edge_type/branch/peer combo
        WITH DISTINCT node_with_edge_element_ids, %(id_func_name)s(node_with_deleted_edge) AS node_with_deleted_edge_element_id
        WITH node_with_edge_element_ids, collect(node_with_deleted_edge_element_id) AS node_with_deleted_edge_element_ids
        RETURN size(node_with_edge_element_ids) = size(node_with_deleted_edge_element_ids) AS all_edges_deleted
    }
    //------------
    // What is the earliest active time for this branch/edge_type/peer_element_id/UUID/labels combination?
    //------------
    CALL (n_uuid, vertex_element_ids, branch, edge_type, peer_element_id) {
        MATCH (n {uuid: n_uuid})%(l_arrow)s-[e {branch: branch, status: "active"}]-%(r_arrow)s(peer)
        WHERE %(id_func_name)s(n) IN vertex_element_ids
        AND type(e) = edge_type
        AND %(id_func_name)s(peer) = peer_element_id
        RETURN e.from AS earliest_active_time
        ORDER BY e.from ASC
        LIMIT 1
    }
    //------------
    // What is the latest deleted time for this branch/edge_type/peer_element_id/UUID/labels combination?
    //------------
    CALL (n_uuid, vertex_element_ids, branch, edge_type, peer_element_id, all_edges_deleted) {
        OPTIONAL MATCH (n {uuid: n_uuid})%(l_arrow)s-[e {branch: branch}]-%(r_arrow)s(peer)
        WHERE all_edges_deleted = TRUE
        AND %(id_func_name)s(n) IN vertex_element_ids
        AND type(e) = edge_type
        AND %(id_func_name)s(peer) = peer_element_id
        RETURN CASE
            WHEN e.status = "active" THEN e.to
            ELSE e.from
        END AS latest_deleted_time
        ORDER BY latest_deleted_time DESC
        LIMIT 1
    }
    // ------------
    // Add the %(direction)s edges to the node we are keeping, if necessary
    // ------------
    CALL (
        n_uuid, vertex_element_ids, element_id_to_keep, branch, edge_type, peer_element_id, all_edges_deleted,
        earliest_active_time, latest_deleted_time
    ) {
        // get the node we are keeping
        MATCH (vertex_to_keep {uuid: n_uuid})
        WHERE %(id_func_name)s(vertex_to_keep) = element_id_to_keep
        // get the peer we are linking to
        MATCH (n {uuid: n_uuid})%(l_arrow)s-[]-%(r_arrow)s(peer)
        WHERE %(id_func_name)s(n) IN vertex_element_ids
        AND %(id_func_name)s(peer) = peer_element_id
        WITH peer, vertex_to_keep
        LIMIT 1
        // ------------
        // mark all other edges for this branch/edge_type/peer combination as to be deleted
        // we will unmark any to keep later
        // ------------
        CALL (n_uuid, branch, peer, vertex_element_ids, edge_type) {
            OPTIONAL MATCH (n {uuid: n_uuid})%(l_arrow)s-[edge_to_delete {branch: branch}]-%(r_arrow)s(peer)
            WHERE %(id_func_name)s(n) IN vertex_element_ids
            AND type(edge_to_delete) = edge_type
            SET edge_to_delete.to_delete = TRUE
        }
        CALL (n_uuid, branch, vertex_element_ids, edge_type, peer) {
            // ------------
            // get the edge to copy
            // ------------
            MATCH (n {uuid: n_uuid})%(l_arrow)s-[e {branch: branch, status: "active"}]-%(r_arrow)s(peer)
            WHERE %(id_func_name)s(n) IN vertex_element_ids
            AND type(e) = edge_type
            RETURN e AS edge_to_copy
            ORDER BY e.from DESC
            LIMIT 1
        }
        %(active_edge_subqueries)s
        // ------------
        // conditionally create the deleted edges
        // ------------
        WITH vertex_to_keep, peer, edge_to_copy
        WHERE all_edges_deleted = TRUE
        %(deleted_edge_subqueries)s
    }
}
        """ % {
            "direction": direction.value,
            "l_arrow": l_arrow,
            "r_arrow": r_arrow,
            "id_func_name": db.get_id_function_name(),
            "active_edge_subqueries": active_edge_subqueries,
            "deleted_edge_subqueries": deleted_edge_subqueries,
            "vertex_label": self.vertex_label,
        }
        return edges_query

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["limit"] = self.limit or 1000
        self.params["offset"] = self.offset or 0
        query_start = """
//------------
// Find vertices with the same labels and UUID
//------------
MATCH (n:%(vertex_label)s)
WITH n.uuid AS node_uuid, count(*) as num_nodes_with_uuid
WHERE num_nodes_with_uuid > 1
WITH DISTINCT node_uuid
ORDER BY node_uuid ASC
MATCH (n:%(vertex_label)s {uuid: node_uuid})
WITH node_uuid, n, %(id_func_name)s(n) AS element_id
ORDER BY node_uuid ASC, element_id ASC
CALL (n) {
    WITH labels(n) AS n_labels
    UNWIND n_labels AS n_label
    WITH n_label
    ORDER BY n_label ASC
    RETURN collect(n_label) AS sorted_labels
}
WITH n.uuid AS n_uuid, sorted_labels, collect(element_id) AS vertex_element_ids
WHERE size(vertex_element_ids) > 1
WITH n_uuid, vertex_element_ids
//------------
// Are there more nodes to process after this query?
//------------
WITH collect([n_uuid, vertex_element_ids]) AS duplicate_details
WITH duplicate_details, size(duplicate_details) > ($offset + $limit) AS more_nodes_to_process
UNWIND duplicate_details AS duplicate_detail
WITH duplicate_detail[0] AS n_uuid, duplicate_detail[1] AS vertex_element_ids, more_nodes_to_process
//------------
// Limit the nodes to process
//------------
SKIP $offset
LIMIT $limit
//------------
// Which node are we going to keep for this UUID/labels combination?
//------------
CALL (vertex_element_ids) {
    UNWIND vertex_element_ids AS element_id
    WITH element_id
    ORDER BY element_id ASC
    RETURN element_id AS element_id_to_keep
    LIMIT 1
}
        """ % {"id_func_name": db.get_id_function_name(), "vertex_label": self.vertex_label}
        self.add_to_query(query_start)

        if self.outbound_edge_types:
            outbound_edges_query = self._build_directed_edges_subquery(
                db=db,
                direction=RelationshipDirection.OUTBOUND,
                edge_types=self.outbound_edge_types,
            )
            self.add_to_query(outbound_edges_query)
        if self.inbound_edge_types:
            inbound_edges_query = self._build_directed_edges_subquery(
                db=db,
                direction=RelationshipDirection.INBOUND,
                edge_types=self.inbound_edge_types,
            )
            self.add_to_query(inbound_edges_query)

        query_end = """
// ------------
// Mark the nodes to delete
// ------------
MATCH (node_to_delete:%(vertex_label)s {uuid: n_uuid})
WHERE %(id_func_name)s(node_to_delete) IN vertex_element_ids
AND %(id_func_name)s(node_to_delete) <> element_id_to_keep
SET node_to_delete.to_delete = TRUE
RETURN more_nodes_to_process
        """ % {"id_func_name": db.get_id_function_name(), "vertex_label": self.vertex_label}
        self.add_to_query(query_end)
        self.return_labels = ["more_nodes_to_process"]


class DeleteDuplicatedEdgesQuery(Query):
    """
    For all Node vertices, find duplicated or overlapping edges of the same status, type, direction, and branch to update and delete
    - one edge will be kept for each pair of nodes and a given status, type, direction, and branch. it will be
        updated to have the earliest "from" and latest "to" times in this group
    - all the other duplicate/overlapping edges will be deleted
    """

    name = "delete_duplicated_edges"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// ------------
// Find vertex pairs that have duplicate edges
// ------------
MATCH (node_with_dup_edges:Node)-[edge]-(peer)
WITH
    node_with_dup_edges,
    type(edge) AS edge_type,
    edge.status AS edge_status,
    edge.branch AS edge_branch,
    peer,
    %(id_func_name)s(startNode(edge)) = %(id_func_name)s(node_with_dup_edges) AS is_outbound
WITH node_with_dup_edges, edge_type, edge_status, edge_branch, peer, is_outbound, count(*) AS num_dup_edges
WHERE num_dup_edges > 1
WITH DISTINCT node_with_dup_edges, edge_type, edge_branch, peer, is_outbound
CALL (node_with_dup_edges, edge_type, edge_branch, peer, is_outbound) {
    // ------------
    // Get the earliest active and deleted edges for this branch
    // ------------
    OPTIONAL MATCH (node_with_dup_edges)-[active_edge {branch: edge_branch, status: "active"}]-(peer)
    WHERE type(active_edge) = edge_type
    AND (%(id_func_name)s(startNode(active_edge)) = %(id_func_name)s(node_with_dup_edges) OR is_outbound = FALSE)
    WITH node_with_dup_edges, edge_type, edge_branch, peer, is_outbound, active_edge
    ORDER BY active_edge.from ASC
    WITH node_with_dup_edges, edge_type, edge_branch, peer, is_outbound, head(collect(active_edge.from)) AS active_from
    OPTIONAL MATCH (node_with_dup_edges)-[deleted_edge {branch: edge_branch, status: "deleted"}]-(peer)
    WHERE %(id_func_name)s(startNode(deleted_edge)) = %(id_func_name)s(node_with_dup_edges) OR is_outbound = FALSE
    WITH node_with_dup_edges, edge_type, edge_branch, peer, is_outbound, active_from, deleted_edge
    ORDER BY deleted_edge.from DESC
    WITH node_with_dup_edges, edge_type, edge_branch, peer, is_outbound, active_from, head(collect(deleted_edge.from)) AS deleted_from
    // ------------
    // ensure one active edge with correct from and to times
    // set the others to be deleted
    // ------------
    CALL (node_with_dup_edges, edge_type, edge_branch, peer, is_outbound, active_from, deleted_from) {
        OPTIONAL MATCH (node_with_dup_edges)-[active_e {branch: edge_branch, status: "active"}]-(peer)
        WHERE type(active_e) = edge_type
        AND (%(id_func_name)s(startNode(active_e)) = %(id_func_name)s(node_with_dup_edges) OR is_outbound = FALSE)
        WITH active_from, deleted_from, collect(active_e) AS active_edges
        WITH active_from, deleted_from, head(active_edges) AS edge_to_keep, tail(active_edges) AS edges_to_delete
        SET edge_to_keep.from = active_from
        SET edge_to_keep.to = deleted_from
        WITH edges_to_delete
        UNWIND edges_to_delete AS edge_to_delete
        SET edge_to_delete.to_delete = TRUE
    }
    // ------------
    // ensure one deleted edge with correct from time, if necessary
    // set the others to be deleted
    // ------------
    CALL (node_with_dup_edges, edge_type, edge_branch, peer, is_outbound, deleted_from) {
        MATCH (node_with_dup_edges)-[deleted_e {branch: edge_branch, status: "deleted"}]-(peer)
        WHERE type(deleted_e) = edge_type
        AND (%(id_func_name)s(startNode(deleted_e)) = %(id_func_name)s(node_with_dup_edges) OR is_outbound = FALSE)
        WITH deleted_from, collect(deleted_e) AS deleted_edges
        WITH deleted_from, head(deleted_edges) AS edge_to_keep, tail(deleted_edges) AS edges_to_delete
        SET edge_to_keep.from = deleted_from
        WITH edges_to_delete
        UNWIND edges_to_delete AS edge_to_delete
        SET edge_to_delete.to_delete = TRUE
    }
}
        """ % {"id_func_name": db.get_id_function_name()}
        self.add_to_query(query)


class DeleteIllegalRelationships(Query):
    """
    Find all Relationship vertices with the same UUID (in a valid database, there are none)
    If any of these Relationships have an IS_RELATED edge to a deleted Node, then delete them
        this includes if an IS_RELATED edge was added on a branch after the Node was deleted on main or -global-
    If any of these Relationships are now only connected to a single Node, then delete them
    """

    name = "delete_illegal_relationships"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// ------------
// Get the default and global branch names
// ------------
MATCH (default_b:Branch)
WHERE default_b.is_default = TRUE
WITH default_b.name AS default_branch
LIMIT 1
MATCH (global_b:Branch)
WHERE global_b.is_global = TRUE
WITH default_branch, global_b.name AS global_branch
LIMIT 1
// ------------
// Find relationships with duplicate UUIDs
// ------------
MATCH (r: Relationship)
WITH default_branch, global_branch, r.uuid AS r_uuid, count(*) AS num_dups
WHERE num_dups > 1
WITH DISTINCT default_branch, global_branch, r_uuid
// ------------
// Find any IS_RELATED edges on the duplicate Relationships that link to deleted Nodes,
//   accounting for if the edge was added on a branch after the Node was deleted on main
// ------------
CALL (default_branch, global_branch, r_uuid) {
    MATCH (:Relationship {uuid: r_uuid})-[is_related:IS_RELATED]-(n:Node)
    CALL (is_related) {
        MATCH (b:Branch {name: is_related.branch})
        RETURN b.branched_from AS edge_branched_from_time
    }
    // ------------
    // If this Node was deleted
    // ------------
    MATCH (n)-[is_part_of:IS_PART_OF]->(:Root)
    WHERE (is_part_of.status = "deleted" OR is_part_of.to IS NOT NULL)
    // ------------
    // before the active IS_RELATED edge's from time, then delete the edge
    // ------------
    WITH is_related, edge_branched_from_time, is_part_of, CASE
        WHEN is_part_of.status = "deleted" THEN is_part_of.from
        ELSE is_part_of.to
    END AS node_deleted_time
    WHERE (is_part_of.branch IN [is_related.branch, global_branch] AND is_related.from > node_deleted_time)
    OR (is_part_of.branch = default_branch AND node_deleted_time < edge_branched_from_time)
    DELETE is_related
}
MATCH (rel:Relationship {uuid: r_uuid})
CALL (rel) {
    OPTIONAL MATCH (rel)-[:IS_RELATED]-(n:Node)
    WITH DISTINCT n
    RETURN count(*) AS num_peers
}
WITH rel
WHERE num_peers < 2
DETACH DELETE rel
        """
        self.add_to_query(query)


class DeleteDuplicateRelationships(Query):
    """
    There can also be leftover duplicate active Relationships that do not have the same UUID.
    They are linked to the same Nodes, have the same Relationship.name, and are on the same branch.
    In this case, we want to DETACH DELETE the later Relationship. We won't lose any information b/c the exact
    same Relationship (maybe with an earlier from time) still exists.
    """

    name = "delete_duplicate_relationships"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (n:Node)
WITH n.uuid AS node_uuid, count(*) as num_nodes_with_uuid
WHERE num_nodes_with_uuid > 1
WITH DISTINCT node_uuid
ORDER BY node_uuid ASC
MATCH (a:Node {uuid: node_uuid})-[e1:IS_RELATED {status: "active"}]-(rel:Relationship)-[e2:IS_RELATED {branch: e1.branch, status: "active"}]-(b:Node)
WHERE a.uuid <> b.uuid
AND e1.to IS NULL
AND e2.to IS NULL
WITH a, rel.name AS rel_name, rel, b, e1.branch AS branch, CASE
    WHEN startNode(e1) = a AND startNode(e2) = rel THEN "out"
    WHEN startNode(e1) = rel AND startNode(e2) = b THEN "in"
    ELSE "bidir"
END AS direction,
CASE
    WHEN e1.from < e2.from THEN e1.from ELSE e2.from
END AS earliest_from
ORDER BY %(id_func_name)s(a), rel_name, %(id_func_name)s(b), direction, branch, earliest_from ASC
WITH a, rel_name, b, direction, branch, collect(rel) AS relationships_list
WHERE size(relationships_list) > 1
WITH a, rel_name, b, direction, branch, tail(relationships_list) AS rels_to_delete
UNWIND rels_to_delete AS rel_to_delete
DETACH DELETE rel_to_delete
        """ % {"id_func_name": db.get_id_function_name()}
        self.add_to_query(query)


class PerformHardDeletes(Query):
    name = "do_hard_deletes"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
CALL () {
    MATCH (n)
    WHERE n.to_delete = TRUE
    DETACH DELETE n
}
CALL () {
    MATCH ()-[e]-()
    WHERE e.to_delete = TRUE
    DELETE e
}
        """
        self.add_to_query(query)


class Migration029(ArbitraryMigration):
    """
    Clean up a variety of bad data created during bugged merges for node kind/inheritance updates

    1. Identify improperly duplicated nodes (ie nodes with the same UUID and the same database labels)
        a. Consolidate edges onto a single duplicated node, making sure that the edges remain active if ANY active path exists
        b. Delete the duplicated edges
        c. Delete the duplicated nodes
    2. Delete duplicated Relationships linked to the de-duplicated node
    3. Delete duplicated edges across the database

    Some of these changes must be batched because there can be a lot of them and the queries can be rather complex
    Some of these queries also require marking nodes and edges as to be deleted (using the `to_delete` property) and then
    deleting them in a separate query
    """

    name: str = "029_duplicates_cleanup"
    minimum_version: int = 28
    limit: int = 100

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()

        return result

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        migration_result = MigrationResult()
        db = migration_input.db
        limit = self.limit
        offset = 0
        more_nodes_to_process = True
        try:
            while more_nodes_to_process:
                log.info(f"Running node duplicates cleanup query {limit=},{offset=}")
                node_cleanup_query = await CleanUpDuplicatedUuidVertices.init(
                    db=db,
                    vertex_label="Node",
                    limit=limit,
                    offset=offset,
                    outbound_edge_types=[
                        DatabaseEdgeType.IS_PART_OF,
                        DatabaseEdgeType.HAS_ATTRIBUTE,
                        DatabaseEdgeType.IS_RELATED,
                        DatabaseEdgeType.IS_RESERVED,
                    ],
                    inbound_edge_types=[
                        DatabaseEdgeType.IS_RELATED,
                        DatabaseEdgeType.HAS_OWNER,
                        DatabaseEdgeType.HAS_SOURCE,
                    ],
                )
                await node_cleanup_query.execute(db=db)
                has_results = False
                for result in node_cleanup_query.get_results():
                    has_results = True
                    more_nodes_to_process = result.get_as_type("more_nodes_to_process", bool)
                offset += limit
                if not has_results or not more_nodes_to_process:
                    break

            hard_delete_query = await PerformHardDeletes.init(db=db)
            await hard_delete_query.execute(db=db)

            duplicate_edge_query = await DeleteDuplicatedEdgesQuery.init(db=db)
            await duplicate_edge_query.execute(db=db)

            hard_delete_query = await PerformHardDeletes.init(db=db)
            await hard_delete_query.execute(db=db)

            illegal_relationships_cleanup_query = await DeleteIllegalRelationships.init(db=db)
            await illegal_relationships_cleanup_query.execute(db=db)

            offset = 0
            more_nodes_to_process = True
            while more_nodes_to_process:
                log.info(f"Running relationship duplicates cleanup query {limit=},{offset=}")
                relationship_cleanup_query = await CleanUpDuplicatedUuidVertices.init(
                    db=db,
                    vertex_label="Relationship",
                    limit=limit,
                    offset=offset,
                    outbound_edge_types=[
                        DatabaseEdgeType.IS_RELATED,
                        DatabaseEdgeType.IS_PROTECTED,
                        DatabaseEdgeType.HAS_OWNER,
                        DatabaseEdgeType.HAS_SOURCE,
                    ],
                    inbound_edge_types=[
                        DatabaseEdgeType.IS_RELATED,
                    ],
                )
                await relationship_cleanup_query.execute(db=db)
                has_results = False
                for result in relationship_cleanup_query.get_results():
                    has_results = True
                    more_nodes_to_process = result.get_as_type("more_nodes_to_process", bool)
                offset += limit
                if not has_results or not more_nodes_to_process:
                    break

            hard_delete_query = await PerformHardDeletes.init(db=db)
            await hard_delete_query.execute(db=db)

            duplicate_relationships_cleanup_query = await DeleteDuplicateRelationships.init(db=db)
            await duplicate_relationships_cleanup_query.execute(db=db)

        except Exception as exc:
            migration_result.errors.append(str(exc))
            return migration_result

        return migration_result



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.branch import Branch
from infrahub.core.migrations.shared import ArbitraryMigration, MigrationInput, MigrationResult
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration051(ArbitraryMigration):
    name: str = "051_subtract_branched_from_microsecond"
    minimum_version: int = 50

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        result = MigrationResult()

        branches = await Branch.get_list(db=db)

        for branch in branches:
            if branch.is_default or branch.is_global:
                continue

            if not branch.branched_from:
                continue

            original_ts = Timestamp(branch.branched_from)
            new_ts = original_ts.subtract(microseconds=1)
            branch.branched_from = new_ts.to_string()

            await branch.save(db=db)

        return result



from __future__ import annotations

import re
import urllib.parse
from typing import TYPE_CHECKING, cast

from fastapi import APIRouter, Depends, HTTPException, Query, Response

from infrahub.api.dependencies import (
    BranchParams,
    get_branch_params,
    get_current_user,
    get_db,
    get_permission_manager,
)
from infrahub.core import registry
from infrahub.core.constants import InfrahubKind, PermissionAction
from infrahub.core.protocols import CoreFileObject
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.exceptions import NodeNotFoundError
from infrahub.permissions import define_object_permission_from_branch

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.permissions import PermissionManager

router = APIRouter(prefix="/files")


def sanitize_filename(filename: str) -> tuple[str, str]:
    """Sanitize filename to prevent header injection attacks.

    Returns a tuple of (ascii_filename, encoded_filename) for use in Content-Disposition header.
    - ascii_filename: Safe ASCII-only filename for the 'filename' parameter
    - encoded_filename: RFC5987 percent-encoded filename for the 'filename*' parameter
    """
    # Strip control characters (CR, LF, NULL, etc.)
    filename = re.sub(r"[\x00-\x1f\x7f]", "", filename)

    # Strip/replace characters that could break the header
    filename = filename.replace('"', "'").replace(";", "_")

    # Truncate to reasonable length (255 is common filesystem limit)
    if len(filename) > 255:
        # Keep extension if present
        if "." in filename:
            ext = filename.rsplit(".", 1)[-1][:10]  # Max 10 char extension
            filename = filename[: 255 - len(ext) - 1] + "." + ext
        else:
            filename = filename[:255]

    # Create ASCII-safe fallback (replace non-ASCII with underscore)
    ascii_filename = filename.encode("ascii", errors="replace").decode("ascii").replace("?", "_")

    # RFC5987 percent-encode for filename* parameter
    encoded_filename = urllib.parse.quote(filename, safe="")

    return ascii_filename, encoded_filename


def build_content_disposition(filename: str, preview: bool = False) -> str:
    """Build a safe Content-Disposition header value.

    https://developer.mozilla.org/docs/Web/HTTP/Reference/Headers/Content-Disposition

    Args:
        filename: The filename to include in the header.
        preview: If True, use 'inline' disposition (display in browser). If False, use 'attachment' (force download).
    """
    ascii_filename, encoded_filename = sanitize_filename(filename)
    disposition = "inline" if preview else "attachment"
    # Use both filename (for older clients) and filename* (for RFC5987 compliant clients)
    return f"{disposition}; filename=\"{ascii_filename}\"; filename*=UTF-8''{encoded_filename}"


def _build_file_response(file_object: CoreFileObject, *, preview: bool = False) -> Response:
    """Build a `Response` for downloading a FileObject's content."""
    return Response(
        content=registry.storage.retrieve_binary(identifier=file_object.storage_id.value),
        media_type=file_object.file_type.value,
        headers={
            "Content-Disposition": build_content_disposition(filename=file_object.file_name.value, preview=preview)
        },
    )


@router.get(
    "/by-hfid/{kind:str}",
    response_class=Response,
    responses={
        200: {
            "description": "File content with Content-Type matching the file's MIME type",
            "content": {"*/*": {"schema": {"type": "string", "format": "binary"}}},
        }
    },
)
async def download_file_object_by_hfid(
    kind: str,
    hfid: list[str] = Query(..., description="HFID component values in order"),
    preview: bool = Query(False, description="Return file for inline display rather than as an attachment"),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    permission_manager: PermissionManager = Depends(get_permission_manager),
    _: AccountSession = Depends(get_current_user),
) -> Response:
    """Download a file by the FileObject node's Human-Friendly ID (HFID).

    Requires `VIEW` permission on the FileObject node.
    Returns the binary file content with `Content-Type` from the node's `file_type` attribute and `Content-Disposition` header with the original
    filename.
    """
    schema = registry.schema.get_node_schema(name=kind, branch=branch_params.branch, duplicate=False)

    if InfrahubKind.FILEOBJECT not in schema.inherit_from:
        raise HTTPException(status_code=400, detail=f"'{kind}' is not a file object")

    permission = define_object_permission_from_branch(
        schema=schema, action=PermissionAction.VIEW, branch_name=branch_params.branch.name
    )
    permission_manager.raise_for_permission(permission=permission)

    node = await registry.manager.get_one_by_hfid(
        db=db, hfid=hfid, kind=kind, branch=branch_params.branch, at=branch_params.at, raise_on_error=True
    )

    return _build_file_response(file_object=cast("CoreFileObject", node), preview=preview)


@router.get(
    "/by-storage-id/{storage_id:str}",
    response_class=Response,
    responses={
        200: {
            "description": "File content with Content-Type matching the file's MIME type",
            "content": {"*/*": {"schema": {"type": "string", "format": "binary"}}},
        }
    },
)
async def download_file_object_by_storage_id(
    storage_id: str,
    preview: bool = Query(False, description="Return file for inline display rather than as an attachment"),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    permission_manager: PermissionManager = Depends(get_permission_manager),
    _: AccountSession = Depends(get_current_user),
) -> Response:
    """Download a file from storage by its storage_id.

    Requires `VIEW` permission on the FileObject node.
    Returns the binary file content with `Content-Type` from the node's `file_type` attribute and `Content-Disposition` header with the original
    filename.
    """
    file_objects = await registry.manager.query(
        db=db,
        schema=CoreFileObject,
        filters={"storage_id__value": storage_id},
        branch=branch_params.branch,
        at=branch_params.at,
        limit=1,
    )
    if not file_objects:
        raise NodeNotFoundError(
            branch_name=branch_params.branch.name, node_type=InfrahubKind.FILEOBJECT, identifier=storage_id
        )

    node = file_objects[0]
    permission = define_object_permission_from_branch(
        schema=node.get_schema(), action=PermissionAction.VIEW, branch_name=branch_params.branch.name
    )
    permission_manager.raise_for_permission(permission=permission)

    return _build_file_response(file_object=node, preview=preview)


@router.get(
    "/{node_id:str}",
    response_class=Response,
    responses={
        200: {
            "description": "File content with Content-Type matching the file's MIME type",
            "content": {"*/*": {"schema": {"type": "string", "format": "binary"}}},
        }
    },
)
async def download_file_object(
    node_id: str,
    preview: bool = Query(False, description="Return file for inline display rather than as an attachment"),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    permission_manager: PermissionManager = Depends(get_permission_manager),
    _: AccountSession = Depends(get_current_user),
) -> Response:
    """Download a file by the FileObject node's UUID.

    Requires `VIEW` permission on the FileObject node.
    Returns the binary file content with `Content-Type` from the node's `file_type` attribute and `Content-Disposition` header with the original
    filename.
    """
    node = await registry.manager.get_one(
        db=db, id=node_id, kind=CoreFileObject, branch=branch_params.branch, at=branch_params.at, raise_on_error=True
    )

    permission = define_object_permission_from_branch(
        schema=node.get_schema(), action=PermissionAction.VIEW, branch_name=branch_params.branch.name
    )
    permission_manager.raise_for_permission(permission=permission)

    return _build_file_response(file_object=node, preview=preview)



from __future__ import annotations

from enum import IntFlag, StrEnum, auto

from infrahub.core.constants import GlobalPermissions


class PermissionDecisionFlag(IntFlag):
    DENY = 1
    ALLOW_DEFAULT = 2
    ALLOW_OTHER = 4
    ALLOW_ALL = ALLOW_DEFAULT | ALLOW_OTHER


class BranchRelativePermissionDecision(StrEnum):
    """This enum is only used to communicate a permission decision relative to a branch."""

    DENY = auto()
    ALLOW = auto()
    ALLOW_DEFAULT = auto()
    ALLOW_OTHER = auto()


GLOBAL_PERMISSION_DENIAL_MESSAGE = {
    GlobalPermissions.EDIT_DEFAULT_BRANCH.value: "You are not allowed to change data in the default branch",
    GlobalPermissions.MERGE_BRANCH.value: "You are not allowed to merge a branch",
    GlobalPermissions.MERGE_PROPOSED_CHANGE.value: "You are not allowed to merge proposed changes",
    GlobalPermissions.REVIEW_PROPOSED_CHANGE.value: "You are not allowed to review proposed changes",
    GlobalPermissions.MANAGE_SCHEMA.value: "You are not allowed to manage the schema",
    GlobalPermissions.MANAGE_ACCOUNTS.value: "You are not allowed to manage user accounts, groups or roles",
    GlobalPermissions.MANAGE_PERMISSIONS.value: "You are not allowed to manage permissions",
    GlobalPermissions.MANAGE_REPOSITORIES.value: "You are not allowed to manage repositories",
    GlobalPermissions.UPDATE_OBJECT_HFID_DISPLAY_LABEL.value: "You are not allowed to update human friendly IDs and display labels ad hoc",
}

GLOBAL_PERMISSION_DESCRIPTION = {
    GlobalPermissions.EDIT_DEFAULT_BRANCH: "Allow a user to change data in the default branch",
    GlobalPermissions.MERGE_BRANCH: "Allow a user to merge branches",
    GlobalPermissions.MERGE_PROPOSED_CHANGE: "Allow a user to merge proposed changes",
    GlobalPermissions.REVIEW_PROPOSED_CHANGE: "Allow a user to approve or reject proposed changes",
    GlobalPermissions.MANAGE_SCHEMA: "Allow a user to manage the schema",
    GlobalPermissions.MANAGE_ACCOUNTS: "Allow a user to manage accounts, account roles and account groups",
    GlobalPermissions.MANAGE_PERMISSIONS: "Allow a user to manage permissions",
    GlobalPermissions.MANAGE_REPOSITORIES: "Allow a user to manage repositories",
    GlobalPermissions.SUPER_ADMIN: "Allow a user to do anything",
    GlobalPermissions.UPDATE_OBJECT_HFID_DISPLAY_LABEL: "Allow a user to update objects' display labels and human friendly IDs ad hoc",
}



import itertools
import re

MATCH_PATTERN = r"(\[[\w,-]*[-,][\w,-]*\])"


def _escape_brackets(s: str) -> str:
    return s.replace("\\[", "__LBRACK__").replace("\\]", "__RBRACK__")


def _unescape_brackets(s: str) -> str:
    return s.replace("__LBRACK__", "[").replace("__RBRACK__", "]")


def _char_range_expand(char_range_str: str) -> list[str]:
    """Expands a string of numbers or single-character letters."""
    expanded_values: list[str] = []
    # Special case: if no dash and no comma, and multiple characters, error if not all alphanumeric
    if "," not in char_range_str and "-" not in char_range_str and len(char_range_str) > 1:
        if not char_range_str.isalnum():
            raise ValueError(f"Invalid non-alphanumeric range: [{char_range_str}]")
        return list(char_range_str)

    for value in char_range_str.split(","):
        if not value:
            # Malformed: empty part in comma-separated list
            return [f"[{char_range_str}]"]
        if "-" in value:
            start_char, end_char = value.split("-", 1)
            if not start_char or not end_char:
                expanded_values.append(f"[{char_range_str}]")
                return expanded_values
            # Check if it's a numeric range
            if start_char.isdigit() and end_char.isdigit():
                start_num = int(start_char)
                end_num = int(end_char)
                step = 1 if start_num <= end_num else -1
                expanded_values.extend(str(i) for i in range(start_num, end_num + step, step))
            # Check if it's an alphabetical range (single character)
            elif len(start_char) == 1 and len(end_char) == 1 and start_char.isalpha() and end_char.isalpha():
                start_ord = ord(start_char)
                end_ord = ord(end_char)
                step = 1 if start_ord <= end_ord else -1
                is_upper = start_char.isupper()
                for i in range(start_ord, end_ord + step, step):
                    char = chr(i)
                    expanded_values.append(char.upper() if is_upper else char)
            else:
                # Mixed or unsupported range type, append as-is
                expanded_values.append(value)
        else:
            # If the value is a single character or valid alphanumeric string, append
            if not value.isalnum():
                raise ValueError(f"Invalid non-alphanumeric value: [{value}]")
            expanded_values.append(value)
    return expanded_values


def _extract_constants(pattern: str, re_compiled: re.Pattern) -> tuple[list[int], list[list[str]]]:
    cartesian_list = []
    interface_constant = [0]
    for match in re_compiled.finditer(pattern):
        interface_constant.extend((match.start(), match.end()))
        cartesian_list.append(_char_range_expand(match.group()[1:-1]))
    return interface_constant, cartesian_list


def _expand_interfaces(pattern: str, interface_constant: list[int], cartesian_list: list[list[str]]) -> list[str]:
    def _pairwise(lst: list[int]) -> list[tuple[int, int]]:
        it = iter(lst)
        return list(zip(it, it, strict=False))

    if interface_constant[-1] < len(pattern):
        interface_constant.append(len(pattern))
    interface_constant_out = _pairwise(interface_constant)
    expanded_interfaces = []
    for element in itertools.product(*cartesian_list):
        current_interface = ""
        for count, item in enumerate(interface_constant_out):
            current_interface += pattern[item[0] : item[1]]
            if count < len(element):
                current_interface += element[count]
        expanded_interfaces.append(_unescape_brackets(current_interface))
    return expanded_interfaces


def range_expansion(interface_pattern: str) -> list[str]:
    """Expand string pattern into a list of strings, supporting both
    number and single-character alphabet ranges. Heavily inspired by
    Netutils interface_range_expansion but adapted to support letters.

    Args:
        interface_pattern: The string pattern that will be parsed to create the list of interfaces.

    Returns:
        Contains the expanded list of interfaces.

    Examples:
        >>> from infrahub_sdk.spec.range_expansion import range_expansion
        >>> range_expansion("Device [A-C]")
        ['Device A', 'Device B', 'Device C']
        >>> range_expansion("FastEthernet[1-2]/0/[10-15]")
        ['FastEthernet1/0/10', 'FastEthernet1/0/11', 'FastEthernet1/0/12',
        'FastEthernet1/0/13', 'FastEthernet1/0/14', 'FastEthernet1/0/15',
        'FastEthernet2/0/10', 'FastEthernet2/0/11', 'FastEthernet2/0/12',
        'FastEthernet2/0/13', 'FastEthernet2/0/14', 'FastEthernet2/0/15']
        >>> range_expansion("GigabitEthernet[a-c]/0/1")
        ['GigabitEtherneta/0/1', 'GigabitEthernetb/0/1', 'GigabitEthernetc/0/1']
        >>> range_expansion("Eth[a,c,e]/0/1")
        ['Etha/0/1', 'Ethc/0/1', 'Ethe/0/1']
    """
    pattern_escaped = _escape_brackets(interface_pattern)
    re_compiled = re.compile(MATCH_PATTERN)
    if not re_compiled.search(pattern_escaped):
        return [_unescape_brackets(pattern_escaped)]
    interface_constant, cartesian_list = _extract_constants(pattern_escaped, re_compiled)
    return _expand_interfaces(pattern_escaped, interface_constant, cartesian_list)



from __future__ import annotations

from typing import TYPE_CHECKING

from prefect.server.events.schemas.events import ReceivedEvent
from prefect.server.events.storage import INTERACTIVE_PAGE_SIZE
from prefect.server.events.storage.database import raw_count_events, read_events

if TYPE_CHECKING:
    from prefect.server.events.filters import EventFilter
    from sqlalchemy.ext.asyncio import AsyncSession


async def query_events(
    session: AsyncSession, filter: EventFilter, page_size: int = INTERACTIVE_PAGE_SIZE, offset: int | None = None
) -> tuple[list[ReceivedEvent], int]:
    count = await raw_count_events(session, filter)  # type: ignore[attr-defined]
    page = await read_events(session, filter, limit=page_size, offset=offset)  # type: ignore[attr-defined]
    events = [ReceivedEvent.model_validate(e, from_attributes=True) for e in page]
    return events, count



from graphene import Field, Int, List, NonNull, ObjectType, String


class TaskLog(ObjectType):
    message = String(required=True)
    severity = String(required=True)
    task_id = String(required=True)
    timestamp = String(required=True)
    id = String(required=False)


class TaskLogNodes(ObjectType):
    node = Field(TaskLog)


class TaskLogEdge(ObjectType):
    edges = List(NonNull(TaskLogNodes), required=True)
    count = Int(required=True)



from infrahub.core.constants import OBJECT_TEMPLATE_RELATIONSHIP_NAME, PROFILES_RELATIONSHIP_NAME

from .generic_schema import GenericSchema
from .node_schema import NodeSchema


class NodeInheritanceHandler:
    def inherit_from_interface(self, node: NodeSchema, interface: GenericSchema) -> None:
        existing_inherited_attributes: dict[str, int] = {
            item.name: idx for idx, item in enumerate(node.attributes) if item.inherited
        }
        existing_inherited_relationships: dict[str, int] = {
            item.name: idx for idx, item in enumerate(node.relationships) if item.inherited
        }

        properties_to_inherit = [
            "human_friendly_id",
            "display_label",
            "display_labels",
            "default_filter",
            "menu_placement",
            "uniqueness_constraints",
            "icon",
            "order_by",
        ]
        for prop_name in properties_to_inherit:
            if getattr(interface, prop_name) and not getattr(node, prop_name):
                setattr(node, prop_name, getattr(interface, prop_name))

        # Build a mapping from source_attribute_id to (index, old_name) for existing inherited attributes
        # This allows us to detect renamed attributes by their source ID
        existing_inherited_attr_by_source_id: dict[str, tuple[int, str]] = {
            item.source_attribute_id: (idx, item.name)
            for idx, item in enumerate(node.attributes)
            if item.inherited and item.source_attribute_id
        }

        # Track renamed attributes to update schema properties (uniqueness_constraints, etc.)
        renamed_attrs: dict[str, str] = {}  # old_name -> new_name

        for attribute in interface.attributes:
            if attribute.name in node.valid_local_names:
                continue

            new_attribute = attribute.duplicate()
            new_attribute.id = None
            new_attribute.inherited = True
            # Store the source generic's attribute ID for rename detection
            new_attribute.source_attribute_id = attribute.id

            # Check if this attribute already exists by source_attribute_id (for rename detection)
            # or by name (for regular updates)
            if attribute.id and attribute.id in existing_inherited_attr_by_source_id:
                # Attribute was renamed in the generic - update the existing one
                item_idx, old_name = existing_inherited_attr_by_source_id[attribute.id]
                if old_name != new_attribute.name:
                    renamed_attrs[old_name] = new_attribute.name
                    # Update tracking structures to prevent stale lookups
                    existing_inherited_attributes.pop(old_name, None)
                    existing_inherited_attributes[new_attribute.name] = item_idx
                node.attributes[item_idx].update_from_generic(other=new_attribute)
                node.attributes[item_idx].name = new_attribute.name  # Update name explicitly
                node.attributes[item_idx].source_attribute_id = new_attribute.source_attribute_id
            elif attribute.name not in existing_inherited_attributes:
                # New attribute added on the generic
                node.attributes.append(new_attribute)
            else:
                # Existing inherited attribute that has not been renamed
                item_idx = existing_inherited_attributes[attribute.name]
                node.attributes[item_idx].update_from_generic(other=new_attribute)
                node.attributes[item_idx].source_attribute_id = new_attribute.source_attribute_id

        for relationship in interface.relationships:
            if relationship.name in node.valid_local_names:
                continue
            if relationship.name in (PROFILES_RELATIONSHIP_NAME, OBJECT_TEMPLATE_RELATIONSHIP_NAME):
                continue

            new_relationship = relationship.duplicate()
            new_relationship.id = None
            new_relationship.inherited = True

            if relationship.name not in existing_inherited_relationships:
                node.relationships.append(new_relationship)
            else:
                item_idx = existing_inherited_relationships[relationship.name]
                node.relationships[item_idx].update_from_generic(other=new_relationship)

        if renamed_attrs:
            self._update_uniqueness_constraints_for_renamed_attributes(node=node, renamed_attrs=renamed_attrs)
            self._update_hfid_for_renamed_attributes(node=node, renamed_attrs=renamed_attrs)
            self._update_order_by_for_renamed_attributes(node=node, renamed_attrs=renamed_attrs)
            self._update_default_filters_for_renamed_attributes(node=node, renamed_attrs=renamed_attrs)
            self._update_display_labels_for_renamed_attributes(node=node, renamed_attrs=renamed_attrs)

    def _update_default_filters_for_renamed_attributes(self, node: NodeSchema, renamed_attrs: dict[str, str]) -> None:
        if not node.default_filter:
            return
        for old_name, new_name in renamed_attrs.items():
            if node.default_filter == old_name or node.default_filter.startswith(f"{old_name}__"):
                node.default_filter = new_name + node.default_filter[len(old_name) :]
                break

    def _update_order_by_for_renamed_attributes(self, node: NodeSchema, renamed_attrs: dict[str, str]) -> None:
        if not node.order_by:
            return
        node.order_by = self._get_updated_renamed_attrs_data(attr_data=node.order_by, renamed_attrs=renamed_attrs)

    def _update_hfid_for_renamed_attributes(self, node: NodeSchema, renamed_attrs: dict[str, str]) -> None:
        if not node.human_friendly_id:
            return
        node.human_friendly_id = self._get_updated_renamed_attrs_data(
            attr_data=node.human_friendly_id, renamed_attrs=renamed_attrs
        )

    def _update_display_labels_for_renamed_attributes(self, node: NodeSchema, renamed_attrs: dict[str, str]) -> None:
        if not node.display_labels:
            return
        node.display_labels = self._get_updated_renamed_attrs_data(
            attr_data=node.display_labels, renamed_attrs=renamed_attrs
        )

    def _update_uniqueness_constraints_for_renamed_attributes(
        self, node: NodeSchema, renamed_attrs: dict[str, str]
    ) -> None:
        if not node.uniqueness_constraints:
            return
        updated_constraints = []
        for constraint_paths in node.uniqueness_constraints:
            updated_constraints.append(
                self._get_updated_renamed_attrs_data(attr_data=constraint_paths, renamed_attrs=renamed_attrs)
            )
        node.uniqueness_constraints = updated_constraints

    def _get_updated_renamed_attrs_data(self, attr_data: list[str], renamed_attrs: dict[str, str]) -> list[str]:
        updated_attrs_data = []
        for data in attr_data:
            updated_attr_data = data
            for old_name, new_name in renamed_attrs.items():
                if data == old_name or data.startswith(f"{old_name}__"):
                    updated_attr_data = new_name + data[len(old_name) :]
                    break
            updated_attrs_data.append(updated_attr_data)
        return updated_attrs_data



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Sequence

from infrahub.core.branch.models import Branch
from infrahub.core.initialization import initialization
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import MigrationInput, MigrationResult
from infrahub.lock import initialize_lock
from infrahub.log import get_logger

from ..shared import InternalSchemaMigration, SchemaMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration038(InternalSchemaMigration):
    """
    Re-run migration 026 after Migration037 updates AttributeValueIndexed vertices correctly so that the call to
    NodeManager.query will work

    If someone is upgrading from 1.2.4 (release before migration 026) or earlier to 1.4.x or later, then migration 026
    fail to find any 0.0.0.0 prefix nodes even if they exist. So we run it again here after migration 037 makes the
    AttributeValueIndexed changes to be sure it completes correctly.
    """

    name: str = "038_prefix_0000_fix"
    minimum_version: int = 37
    migrations: Sequence[SchemaMigration] = []

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        # load schemas from database into registry
        initialize_lock()
        await initialization(db=db)

        for branch in await Branch.get_list(db=db):
            prefix_0000s = await NodeManager.query(
                db=db, schema="BuiltinIPPrefix", branch=branch, filters={"prefix__values": ["0.0.0.0/0", "::/0"]}
            )
            if not prefix_0000s:
                continue
            ipam_reconciler = IpamReconciler(db=db, branch=branch)
            for prefix in prefix_0000s:
                ip_namespace = await prefix.ip_namespace.get_peer(db=db)
                ip_network = ipaddress.ip_network(prefix.prefix.value)
                await ipam_reconciler.reconcile(
                    ip_value=ip_network,
                    namespace=ip_namespace,
                    node_uuid=prefix.get_id(),
                    at=at,
                )

        return MigrationResult()



from __future__ import annotations

import hashlib
from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.core.schema import SchemaAttributePath


@dataclass
class HFIDDefinition:
    hfid: list[str]
    attributes: set[str] = field(default_factory=set)
    relationships: set[str] = field(default_factory=set)
    relationship_fields: dict[str, set[str]] = field(default_factory=dict)
    filter_key: str = "ids"

    @property
    def fields(self) -> list[str]:
        return sorted(list(self.attributes) + list(self.relationships))

    @property
    def has_related_components(self) -> bool:
        """Indicate if the associated template use variables from relationships"""
        return len(self.relationships) > 0

    def get_hash(self) -> str:
        return hashlib.md5("::".join(self.hfid).encode(), usedforsecurity=False).hexdigest()


@dataclass
class RelationshipIdentifier:
    kind: str
    hfid: list[str]
    filter_key: str

    def __hash__(self) -> int:
        return hash(f"{self.kind}::{'::'.join(self.hfid)}::{self.filter_key}")


@dataclass
class RelationshipTriggers:
    attributes: dict[str, set[RelationshipIdentifier]] = field(default_factory=dict)


class HFIDs:
    def __init__(
        self,
        node_level_hfids: dict[str, HFIDDefinition] | None = None,
        relationship_triggers: dict[str, RelationshipTriggers] | None = None,
    ) -> None:
        self._node_level_hfids: dict[str, HFIDDefinition] = node_level_hfids or {}
        self._relationship_triggers: dict[str, RelationshipTriggers] = relationship_triggers or {}

    def duplicate(self) -> HFIDs:
        return self.__class__(
            node_level_hfids=deepcopy(self._node_level_hfids),
            relationship_triggers=deepcopy(self._relationship_triggers),
        )

    def register_hfid_schema_path(self, kind: str, schema_path: SchemaAttributePath, hfid: list[str]) -> None:
        """Register HFID using the schema path of each impacted schema path in use."""
        if kind not in self._node_level_hfids:
            self._node_level_hfids[kind] = HFIDDefinition(hfid=hfid)
        if schema_path.is_type_attribute:
            self._node_level_hfids[kind].attributes.add(schema_path.active_attribute_schema.name)
        elif schema_path.is_type_relationship and schema_path.related_schema:
            self._node_level_hfids[kind].relationships.add(schema_path.active_relationship_schema.name)
            if schema_path.active_relationship_schema.name not in self._node_level_hfids[kind].relationship_fields:
                self._node_level_hfids[kind].relationship_fields[schema_path.active_relationship_schema.name] = set()
            self._node_level_hfids[kind].relationship_fields[schema_path.active_relationship_schema.name].add(
                schema_path.active_attribute_schema.name
            )
            if schema_path.related_schema.kind not in self._relationship_triggers:
                self._relationship_triggers[schema_path.related_schema.kind] = RelationshipTriggers()
            if (
                schema_path.active_attribute_schema.name
                not in self._relationship_triggers[schema_path.related_schema.kind].attributes
            ):
                self._relationship_triggers[schema_path.related_schema.kind].attributes[
                    schema_path.active_attribute_schema.name
                ] = set()
            self._relationship_triggers[schema_path.related_schema.kind].attributes[
                schema_path.active_attribute_schema.name
            ].add(
                RelationshipIdentifier(
                    kind=kind, filter_key=f"{schema_path.active_relationship_schema.name}__ids", hfid=hfid
                )
            )

    def targets_node(self, kind: str) -> bool:
        """Indicates if there is a human_friendly_id defined for the targeted node"""
        return kind in self._node_level_hfids

    def get_node_definition(self, kind: str) -> HFIDDefinition:
        """Return node kinds together with their template definitions."""
        return self._node_level_hfids[kind]

    def get_template_nodes(self) -> dict[str, HFIDDefinition]:
        """Return node kinds together with their template definitions."""
        return self._node_level_hfids

    def get_related_trigger_nodes(self) -> dict[str, RelationshipTriggers]:
        """Return node kinds that other nodes use within their templates for display_labels."""
        return self._relationship_triggers

    def get_related_definition(self, related_kind: str, target_kind: str) -> HFIDDefinition:
        relationship_trigger = self._relationship_triggers[related_kind]
        for applicable_kinds in relationship_trigger.attributes.values():
            for relationship_identifier in applicable_kinds:
                if target_kind == relationship_identifier.kind:
                    template_label = self.get_node_definition(kind=target_kind)
                    template_label.filter_key = relationship_identifier.filter_key
                    return template_label

        raise ValueError(
            f"Unable to find registered template for {target_kind} registered on related node {related_kind}"
        )



from __future__ import annotations

from infrahub_sdk.graphql import Query
from prefect import task
from prefect.cache_policies import NONE

from infrahub.core.constants import InfrahubKind
from infrahub.core.registry import registry
from infrahub.database import InfrahubDatabase  # noqa: TC001  needed for prefect flow
from infrahub.graphql.execution import execute_graphql_query
from infrahub.graphql.initialization import prepare_graphql_params

from .models import ActionTriggerRuleTriggerDefinition
from .parsers import parse_trigger_rule_response


@task(
    name="gather-trigger-action-rules",
    cache_policy=NONE,
)
async def gather_trigger_action_rules(db: InfrahubDatabase) -> list[ActionTriggerRuleTriggerDefinition]:
    trigger_query = Query(
        name=InfrahubKind.TRIGGERRULE,
        query={
            InfrahubKind.TRIGGERRULE: {
                "edges": {
                    "node": {
                        "__typename": None,
                        "id": None,
                        "name": {"value": None},
                        "branch_scope": {"value": None},
                        "active": {"value": None},
                        "... on CoreNodeTriggerRule": {
                            "node_kind": {"value": None},
                            "mutation_action": {"value": None},
                            "matches": {
                                "edges": {
                                    "node": {
                                        "__typename": None,
                                        "id": None,
                                        "... on CoreNodeTriggerAttributeMatch": {
                                            "attribute_name": {"value": None},
                                            "value": {"value": None},
                                            "value_previous": {"value": None},
                                            "value_match": {"value": None},
                                        },
                                        "... on CoreNodeTriggerRelationshipMatch": {
                                            "relationship_name": {"value": None},
                                            "modification_type": {"value": None},
                                            "peer": {"value": None},
                                        },
                                    }
                                }
                            },
                        },
                        "... on CoreGroupTriggerRule": {
                            "member_update": {"value": None},
                            "group": {
                                "node": {
                                    "id": None,
                                    "__typename": None,
                                }
                            },
                        },
                        "action": {
                            "node": {
                                "__typename": None,
                                "id": None,
                                "name": {"value": None},
                                "... on CoreGroupAction": {
                                    "member_action": {"value": None},
                                    "group": {
                                        "node": {
                                            "id": None,
                                            "__typename": None,
                                        }
                                    },
                                },
                                "... on CoreGeneratorAction": {
                                    "generator": {
                                        "node": {
                                            "__typename": None,
                                            "id": None,
                                        }
                                    },
                                },
                            }
                        },
                    }
                },
            }
        },
    )
    gql_params = await prepare_graphql_params(
        db=db,
        branch=registry.default_branch,
    )
    response = await execute_graphql_query(
        schema=gql_params.schema,
        source=trigger_query.render(),
        context_value=gql_params.context,
        root_value=None,
        variable_values={},
    )

    data = response.data or {}
    trigger_rules = parse_trigger_rule_response(data)
    trigger_candidates = [
        ActionTriggerRuleTriggerDefinition.from_trigger_rule(trigger_rule=trigger_rule)
        for trigger_rule in trigger_rules
        if trigger_rule.active
    ]

    return [trigger_rule for trigger_rule in trigger_candidates if trigger_rule]



from infrahub.core.constants import (
    AccountStatus,
    AccountType,
    BranchSupportType,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality

from ...attribute_schema import AttributeSchema as Attr
from ...dropdown import DropdownChoice
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_account = NodeSchema(
    name="Account",
    namespace="Core",
    description="User Account for Infrahub",
    include_in_menu=False,
    label="Account",
    icon="mdi:account",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["label__value"],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE, InfrahubKind.GENERICACCOUNT],
)

core_account_token = NodeSchema(
    name="AccountToken",
    namespace="Internal",
    description="Token for User Account",
    include_in_menu=False,
    label="Account Token",
    default_filter="token__value",
    display_labels=["token__value"],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    uniqueness_constraints=[["token__value"]],
    documentation="/topics/auth",
    attributes=[
        Attr(name="name", kind="Text", optional=True),
        Attr(name="token", kind="Text", description="The authentication token value", unique=True),
        Attr(name="expiration", kind="DateTime", description="Date and time when the token expires", optional=True),
    ],
    relationships=[
        Rel(
            name="account",
            peer=InfrahubKind.GENERICACCOUNT,
            optional=False,
            cardinality=Cardinality.ONE,
            identifier="account__token",
        ),
    ],
)

core_password_credential = NodeSchema(
    name="PasswordCredential",
    namespace="Core",
    description="Username/Password based credential",
    include_in_menu=False,
    label="Username / Password",
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.CREDENTIAL],
    attributes=[
        Attr(name="username", kind="Text", optional=True, branch=BranchSupportType.AGNOSTIC, order_weight=6000),
        Attr(name="password", kind="Password", optional=True, branch=BranchSupportType.AGNOSTIC, order_weight=7000),
    ],
)

core_refresh_token = NodeSchema(
    name="RefreshToken",
    namespace="Internal",
    description="Refresh Token",
    include_in_menu=False,
    label="Refresh Token",
    display_labels=[],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    attributes=[
        Attr(
            name="expiration",
            kind="DateTime",
            description="Date and time when the refresh token expires",
            optional=False,
        ),
    ],
    relationships=[
        Rel(
            name="account",
            peer=InfrahubKind.GENERICACCOUNT,
            optional=False,
            cardinality=Cardinality.ONE,
            identifier="account__refreshtoken",
        ),
    ],
)

core_credential = GenericSchema(
    name="Credential",
    namespace="Core",
    description="A credential that could be referenced to access external services.",
    include_in_menu=False,
    label="Credential",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["label__value"],
    icon="mdi:key-variant",
    human_friendly_id=["name__value"],
    branch=BranchSupportType.AGNOSTIC,
    uniqueness_constraints=[["name__value"]],
    documentation="/topics/auth",
    attributes=[
        Attr(name="name", kind="Text", unique=True, order_weight=1000),
        Attr(name="label", kind="Text", optional=True, order_weight=2000),
        Attr(name="description", kind="Text", optional=True, order_weight=3000),
    ],
)

core_generic_account = GenericSchema(
    name="GenericAccount",
    namespace="Core",
    description="User Account for Infrahub",
    include_in_menu=False,
    label="Account",
    icon="mdi:account",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["label__value"],
    human_friendly_id=["name__value"],
    branch=BranchSupportType.AGNOSTIC,
    documentation="/topics/auth",
    uniqueness_constraints=[["name__value"]],
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="password", kind="HashedPassword", unique=False),
        Attr(name="label", kind="Text", optional=True),
        Attr(name="description", kind="Text", optional=True),
        Attr(
            name="account_type",
            kind="Text",
            description="Type of account (user, script, etc.)",
            default_value=AccountType.USER.value,
            enum=AccountType.available_types(),
        ),
        Attr(
            name="status",
            kind="Dropdown",
            choices=[
                DropdownChoice(
                    name=AccountStatus.ACTIVE.value,
                    label="Active",
                    description="Account is allowed to login",
                    color="#52be80",
                ),
                DropdownChoice(
                    name=AccountStatus.INACTIVE.value,
                    label="Inactive",
                    description="Account is not allowed to login",
                    color="#e74c3c",
                ),
            ],
            default_value=AccountStatus.ACTIVE.value,
        ),
    ],
    relationships=[Rel(name="tokens", peer=InfrahubKind.ACCOUNTTOKEN, optional=True, cardinality=Cardinality.MANY)],
)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Sequence

from infrahub_sdk.utils import str_to_bool

from infrahub.core.constants import DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.diff.model.path import ConflictSelection
from infrahub.exceptions import SchemaNotFoundError

from .models import (
    AttributeChangelog,
    NodeChangelog,
    RelationshipCardinalityManyChangelog,
    RelationshipCardinalityOneChangelog,
    RelationshipPeerChangelog,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.diff.model.path import (
        EnrichedDiffAttribute,
        EnrichedDiffNode,
        EnrichedDiffProperty,
        EnrichedDiffRelationship,
        EnrichedDiffRoot,
    )
    from infrahub.core.models import SchemaUpdateMigrationInfo
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase


@dataclass
class NodeInDiff:
    node_id: str
    kind: str
    label: str


class DiffChangelogCollector:
    def __init__(
        self,
        diff: EnrichedDiffRoot,
        branch: Branch,
        db: InfrahubDatabase,
        migration_tracker: MigrationTracker | None = None,
    ) -> None:
        self._diff = diff
        self._branch = branch
        self._db = db
        self._diff_nodes: dict[str, NodeInDiff]
        self.migration = migration_tracker or MigrationTracker()

    def _populate_diff_nodes(self) -> None:
        self._diff_nodes = {
            node.uuid: NodeInDiff(node_id=node.uuid, kind=node.kind, label=node.label) for node in self._diff.nodes
        }

    def get_node(self, node_id: str) -> NodeInDiff:
        return self._diff_nodes[node_id]

    def get_peer_kind(self, peer_id: str, node_kind: str, relationship_name: str) -> str:
        """If the peer kind doesn't exist in the diff use the peer kind from the schema"""
        try:
            return self.get_node(node_id=peer_id).kind
        except KeyError:
            schema = self._db.schema.get(node_kind, branch=self._branch, duplicate=False)
            rel_schema = schema.get_relationship(name=relationship_name)
            return rel_schema.peer

    def _process_node(self, node: EnrichedDiffNode) -> NodeChangelog:
        node_changelog = NodeChangelog(node_id=node.uuid, node_kind=node.kind, display_label=node.label)
        try:
            schema = self._db.schema.get(node_changelog.node_kind, branch=self._branch, duplicate=False)
        except SchemaNotFoundError:
            # if the schema has been deleted on self._branch
            schema = None
        for attribute in node.attributes:
            self._process_node_attribute(node=node_changelog, attribute=attribute, schema=schema)

        for relationship in node.relationships:
            self._process_node_relationship(node=node_changelog, relationship=relationship)

        return node_changelog

    def _process_node_attribute(
        self, node: NodeChangelog, attribute: EnrichedDiffAttribute, schema: MainSchemaTypes | None
    ) -> None:
        if schema is None:
            attribute_kind = "n/a"
        else:
            try:
                schema_attribute = schema.get_attribute(name=attribute.name)
                attribute_kind = schema_attribute.kind
            except ValueError:
                # This would currently happen if there has been a schema migration as part of the merge
                # then we don't have access to the attribute kind
                attribute_kind = "n/a"

        changelog_attribute = AttributeChangelog(
            name=self.migration.get_attribute_name(node=node, attribute=attribute), kind=attribute_kind
        )
        for attr_property in attribute.properties:
            match attr_property.property_type:
                case DatabaseEdgeType.HAS_VALUE:
                    # TODO deserialize correct value type from string
                    if _keep_branch_update(diff_property=attr_property):
                        changelog_attribute.set_value(value=attr_property.new_value)
                        changelog_attribute.set_value_previous(value=attr_property.previous_value)
                case DatabaseEdgeType.IS_PROTECTED:
                    if _keep_branch_update(diff_property=attr_property):
                        changelog_attribute.add_property(
                            name="is_protected",
                            value_current=self._convert_string_boolean_value(value=attr_property.new_value),
                            value_previous=self._convert_string_boolean_value(value=attr_property.previous_value),
                        )
                case DatabaseEdgeType.HAS_SOURCE:
                    if _keep_branch_update(diff_property=attr_property):
                        changelog_attribute.add_property(
                            name="source",
                            value_current=attr_property.new_value,
                            value_previous=attr_property.previous_value,
                        )
                case DatabaseEdgeType.HAS_OWNER:
                    if _keep_branch_update(diff_property=attr_property):
                        changelog_attribute.add_property(
                            name="owner",
                            value_current=attr_property.new_value,
                            value_previous=attr_property.previous_value,
                        )

        node.add_attribute(attribute=changelog_attribute)

    def _process_node_relationship(self, node: NodeChangelog, relationship: EnrichedDiffRelationship) -> None:
        match relationship.cardinality:
            case RelationshipCardinality.ONE:
                self._process_node_cardinality_one_relationship(node=node, relationship=relationship)

            case RelationshipCardinality.MANY:
                self._process_node_cardinality_many_relationship(node=node, relationship=relationship)

    def _process_node_cardinality_one_relationship(
        self, node: NodeChangelog, relationship: EnrichedDiffRelationship
    ) -> None:
        changelog_rel = RelationshipCardinalityOneChangelog(name=relationship.name)
        for entry in relationship.relationships:
            for rel_prop in entry.properties:
                match rel_prop.property_type:
                    case DatabaseEdgeType.IS_RELATED:
                        if rel_prop.new_value:
                            changelog_rel.peer_id = rel_prop.new_value
                            changelog_rel.peer_kind = self.get_peer_kind(
                                peer_id=rel_prop.new_value,
                                node_kind=node.node_kind,
                                relationship_name=relationship.name,
                            )
                        if rel_prop.previous_value:
                            changelog_rel.peer_id_previous = rel_prop.previous_value
                            changelog_rel.peer_kind_previous = self.get_peer_kind(
                                peer_id=rel_prop.previous_value,
                                node_kind=node.node_kind,
                                relationship_name=relationship.name,
                            )
                    case DatabaseEdgeType.IS_PROTECTED:
                        changelog_rel.add_property(
                            name="is_protected",
                            value_current=self._convert_string_boolean_value(value=rel_prop.new_value),
                            value_previous=self._convert_string_boolean_value(value=rel_prop.previous_value),
                        )
                    case DatabaseEdgeType.HAS_OWNER:
                        changelog_rel.add_property(
                            name="owner",
                            value_current=rel_prop.new_value,
                            value_previous=rel_prop.previous_value,
                        )
                    case DatabaseEdgeType.HAS_SOURCE:
                        changelog_rel.add_property(
                            name="source",
                            value_current=rel_prop.new_value,
                            value_previous=rel_prop.previous_value,
                        )

        node.add_relationship(relationship_changelog=changelog_rel)

    def _convert_string_boolean_value(self, value: str | None) -> bool | None:
        """Convert string based boolean for is_protected."""
        if value is not None:
            return str_to_bool(value)

        return None

    def _process_node_cardinality_many_relationship(
        self, node: NodeChangelog, relationship: EnrichedDiffRelationship
    ) -> None:
        changelog_rel = RelationshipCardinalityManyChangelog(name=relationship.name)
        for peer in relationship.relationships:
            peer_log = RelationshipPeerChangelog(
                peer_id=peer.peer_id,
                peer_kind=self.get_peer_kind(
                    peer_id=peer.peer_id, node_kind=node.node_kind, relationship_name=relationship.name
                ),
                peer_status=peer.action,
            )
            for peer_prop in peer.properties:
                match peer_prop.property_type:
                    case DatabaseEdgeType.IS_PROTECTED:
                        peer_log.add_property(
                            name="is_protected",
                            value_current=self._convert_string_boolean_value(value=peer_prop.new_value),
                            value_previous=self._convert_string_boolean_value(value=peer_prop.previous_value),
                        )
                    case DatabaseEdgeType.HAS_OWNER:
                        peer_log.add_property(
                            name="owner",
                            value_current=peer_prop.new_value,
                            value_previous=peer_prop.previous_value,
                        )
                    case DatabaseEdgeType.HAS_SOURCE:
                        peer_log.add_property(
                            name="source",
                            value_current=peer_prop.new_value,
                            value_previous=peer_prop.previous_value,
                        )

            changelog_rel.peers.append(peer_log)

        node.add_relationship(relationship_changelog=changelog_rel)

    def collect_changelogs(self) -> Sequence[tuple[DiffAction, NodeChangelog]]:
        self._populate_diff_nodes()
        changelogs = [
            (node.action, self._process_node(node=node))
            for node in self._diff.nodes
            if node.action != DiffAction.UNCHANGED
        ]
        return [(action, node_changelog) for action, node_changelog in changelogs if node_changelog.has_changes]


def _keep_branch_update(diff_property: EnrichedDiffProperty) -> bool:
    if diff_property.conflict and diff_property.conflict.selected_branch == ConflictSelection.BASE_BRANCH:
        return False
    return True


class MigrationTracker:
    """Keeps track of schema updates that happened as part of a migration"""

    def __init__(self, migrations: list[SchemaUpdateMigrationInfo] | None = None) -> None:
        # A dictionary of Node kind, previous attribute name and new attribute
        # {"TestPerson": {"old_attribute_name": "new_attribute_name"}}
        self._migrations_attribute_map: dict[str, dict[str, str]] = {}

        migrations = migrations or []
        for migration in migrations:
            if migration.migration_name == "attribute.name.update":
                if migration.path.schema_kind not in self._migrations_attribute_map:
                    self._migrations_attribute_map[migration.path.schema_kind] = {}
                if migration.path.property_name and migration.path.field_name:
                    self._migrations_attribute_map[migration.path.schema_kind][migration.path.property_name] = (
                        migration.path.field_name
                    )

    def get_attribute_name(self, node: NodeChangelog, attribute: EnrichedDiffAttribute) -> str:
        """Return the current name of the requested attribute"""
        if node.node_kind not in self._migrations_attribute_map:
            return attribute.name
        if attribute.name not in self._migrations_attribute_map[node.node_kind]:
            return attribute.name

        return self._migrations_attribute_map[node.node_kind][attribute.name]



from ..models import EdgeToDelete, EdgeToUpdate, PatchPlan
from .base import PatchQuery


class DeleteDuplicatedEdgesPatchQuery(PatchQuery):
    """
    For all Node vertices, find duplicated or overlapping edges of the same status, type, and branch to update and delete
    - one edge will be kept for each pair of nodes and a given status, type, and branch. it will be
        updated to have the earliest "from" and "to" times in this group
    - all the other duplicate/overlapping edges will be deleted
    """

    @property
    def name(self) -> str:
        return "delete-duplicated-edges"

    async def plan(self) -> PatchPlan:
        query = """
// ------------
// Find vertex pairs that have duplicate edges
// ------------
MATCH (node_with_dup_edges:Node)-[edge]-(peer)
WITH node_with_dup_edges, type(edge) AS edge_type, edge.status AS edge_status, edge.branch AS edge_branch, peer, count(*) AS num_dup_edges
WHERE num_dup_edges > 1
WITH DISTINCT node_with_dup_edges, edge_type, edge_branch, peer
CALL (node_with_dup_edges, edge_type, edge_branch, peer) {
    // ------------
    // Get the earliest active and deleted edges for this branch
    // ------------
    OPTIONAL MATCH (node_with_dup_edges)-[active_edge {branch: edge_branch, status: "active"}]->(peer)
    WHERE type(active_edge) = edge_type
    WITH node_with_dup_edges, edge_type, edge_branch, peer, active_edge
    ORDER BY active_edge.from ASC
    WITH node_with_dup_edges, edge_type, edge_branch, peer, head(collect(active_edge.from)) AS active_from
    OPTIONAL MATCH (node_with_dup_edges)-[deleted_edge {branch: edge_branch, status: "deleted"}]-(peer)
    WITH node_with_dup_edges, edge_type, edge_branch, peer, active_from, deleted_edge
    ORDER BY deleted_edge.from ASC
    WITH node_with_dup_edges, edge_type, edge_branch, peer, active_from, head(collect(deleted_edge.from)) AS deleted_from
    // ------------
    // Plan one active edge update with correct from and to times
    // ------------
    CALL (node_with_dup_edges, edge_type, edge_branch, peer, active_from, deleted_from) {
        OPTIONAL MATCH (node_with_dup_edges)-[active_e {branch: edge_branch, status: "active"}]->(peer)
        WHERE type(active_e) = edge_type
        WITH node_with_dup_edges, edge_type, edge_branch, peer, active_from, deleted_from, active_e
        ORDER BY %(id_func_name)s(active_e)
        LIMIT 1
        WITH active_e, properties(active_e) AS before_props, {from: active_from, to: deleted_from} AS prop_updates
        RETURN CASE
            WHEN active_e IS NOT NULL THEN [
                {
                    db_id: %(id_func_name)s(active_e), before_props: before_props, prop_updates: prop_updates
                }
            ]
            ELSE []
        END AS active_edges_to_update
    }
    // ------------
    // Plan deletes for all the other active edges of this type on this branch
    // ------------
    CALL (node_with_dup_edges, edge_type, edge_branch, peer) {
        OPTIONAL MATCH (node_with_dup_edges)-[active_e {branch: edge_branch, status: "active"}]->(peer)
        WHERE type(active_e) = edge_type
        WITH node_with_dup_edges, peer, active_e
        ORDER BY %(id_func_name)s(active_e)
        SKIP 1
        WITH CASE
            WHEN active_e IS NOT NULL THEN {
                db_id: %(id_func_name)s(active_e),
                from_id: %(id_func_name)s(startNode(active_e)),
                to_id: %(id_func_name)s(endNode(active_e)),
                edge_type: type(active_e),
                before_props: properties(active_e)
            }
            ELSE NULL
        END AS serialized_edge
        RETURN collect(serialized_edge) AS active_edges_to_delete
    }
    // ------------
    // Plan one deleted edge update with correct from time
    // ------------
    CALL (node_with_dup_edges, edge_type, edge_branch, peer, deleted_from) {
        OPTIONAL MATCH (node_with_dup_edges)-[deleted_e {branch: edge_branch, status: "deleted"}]->(peer)
        WHERE type(deleted_e) = edge_type
        WITH node_with_dup_edges, edge_type, edge_branch, peer, deleted_from, deleted_e
        ORDER BY %(id_func_name)s(deleted_e)
        LIMIT 1
        WITH deleted_e, properties(deleted_e) AS before_props, {from: deleted_from} AS prop_updates
        RETURN CASE
            WHEN deleted_e IS NOT NULL THEN [
                {
                    db_id: %(id_func_name)s(deleted_e), before_props: before_props, prop_updates: prop_updates
                }
            ]
            ELSE []
        END AS deleted_edges_to_update
    }
    // ------------
    // Plan deletes for all the other deleted edges of this type on this branch
    // ------------
    CALL (node_with_dup_edges, edge_type, edge_branch, peer) {
        OPTIONAL MATCH (node_with_dup_edges)-[deleted_e {branch: edge_branch, status: "deleted"}]->(peer)
        WHERE type(deleted_e) = edge_type
        WITH node_with_dup_edges, peer, deleted_e
        ORDER BY %(id_func_name)s(deleted_e)
        SKIP 1
        WITH CASE
            WHEN deleted_e IS NOT NULL THEN {
                db_id: %(id_func_name)s(deleted_e),
                from_id: %(id_func_name)s(startNode(deleted_e)),
                to_id: %(id_func_name)s(endNode(deleted_e)),
                edge_type: type(deleted_e),
                before_props: properties(deleted_e)
            }
            ELSE NULL
        END AS serialized_edge

        RETURN collect(serialized_edge) AS deleted_edges_to_delete
    }
    RETURN
        active_edges_to_update + deleted_edges_to_update AS edges_to_update,
        active_edges_to_delete + deleted_edges_to_delete AS edges_to_delete
}
RETURN edges_to_update, edges_to_delete
        """ % {"id_func_name": self.db.get_id_function_name()}
        results = await self.db.execute_query(query=query)
        edges_to_delete: list[EdgeToDelete] = []
        edges_to_update: list[EdgeToUpdate] = []
        for result in results:
            for serial_edge_to_delete in result.get("edges_to_delete"):
                edge_to_delete = EdgeToDelete(**serial_edge_to_delete)
                edges_to_delete.append(edge_to_delete)
            for serial_edge_to_update in result.get("edges_to_update"):
                prop_updates = serial_edge_to_update["prop_updates"]
                if prop_updates:
                    serial_edge_to_update["after_props"] = serial_edge_to_update["before_props"] | prop_updates
                    del serial_edge_to_update["prop_updates"]
                edge_to_update = EdgeToUpdate(**serial_edge_to_update)
                edges_to_update.append(edge_to_update)
        return PatchPlan(
            name=self.name,
            edges_to_delete=edges_to_delete,
            edges_to_update=edges_to_update,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class DeduplicateRelationshipVerticesQuery(Query):
    """
    For each group of duplicate Relationships with the same UUID, delete any Relationship that meets the following criteria:
    - is linked to a deleted node (only if the delete time is before the Relationship's from time)
    - is linked to a node on an incorrect branch (ie Relationship added on main, but Node is on a branch)
    """

    name = "deduplicate_relationship_vertices"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (root:Root)
WITH root.default_branch AS default_branch_name
// ------------
// Find all Relationship vertices with duplicate UUIDs
// ------------
MATCH (r:Relationship)
WITH r.uuid AS r_uuid, default_branch_name, count(*) AS num_dups
WHERE num_dups > 1
WITH DISTINCT r_uuid, default_branch_name
// ------------
// get the branched_from time for each relationship edge and node
// ------------
MATCH (rel:Relationship {uuid: r_uuid})
CALL (rel) {
    MATCH (rel)-[is_rel_e:IS_RELATED {status: "active"}]-(n:Node)
    MATCH (rel_branch:Branch {name: is_rel_e.branch})
    RETURN is_rel_e, rel_branch.branched_from AS rel_branched_from, n
}
// ------------
// for each IS_RELATED edge of the relationship, check the IS_PART_OF edges of the Node vertex
// to determine if this side of the relationship is legal
// ------------
CALL (n, is_rel_e, rel_branched_from, default_branch_name) {
    OPTIONAL MATCH (n)-[is_part_of_e:IS_PART_OF {status: "active"}]->(:Root)
    WHERE (
        // the Node's create time must precede the Relationship's create time
        is_part_of_e.from <= is_rel_e.from AND (is_part_of_e.to >= is_rel_e.from OR is_part_of_e.to IS NULL)
        // the Node must have been created on a branch of equal or lesser depth than the Relationship
        AND is_part_of_e.branch_level <= is_rel_e.branch_level
        // if the Node and Relationships were created on branch_level = 2, then they must be on the same branch
        AND (
            is_part_of_e.branch_level = 1
            OR is_part_of_e.branch = is_rel_e.branch
        )
        // if the Node was created on the default branch, and the Relationship was created on a branch,
        // then the Node must have been created after the branched_from time of the Relationship's branch
        AND (
            is_part_of_e.branch <> default_branch_name
            OR is_rel_e.branch_level = 1
            OR is_part_of_e.from <= rel_branched_from
        )
    )
    WITH is_part_of_e IS NOT NULL AS is_legal
    ORDER BY is_legal DESC
    RETURN is_legal
    LIMIT 1
}
WITH rel, is_legal
ORDER BY rel, is_legal ASC
WITH rel, head(collect(is_legal)) AS is_legal
WHERE is_legal = false
DETACH DELETE rel
        """
        self.add_to_query(query)


class Migration033(GraphMigration):
    """
    Identifies duplicate Relationship vertices that have the same UUID property. Deletes any duplicates that
    are linked to deleted nodes or nodes on in incorrect branch.
    """

    name: str = "033_deduplicate_relationship_vertices"
    minimum_version: int = 32
    queries: Sequence[type[Query]] = [DeduplicateRelationshipVerticesQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

from graphene import Boolean, DateTime, Field, InputObjectType, ObjectType

from infrahub.graphql.types.enums import InfrahubOrderDirection


class InfrahubNodeMetadataOrder(InputObjectType):
    created_at = Field(InfrahubOrderDirection, required=False, description="Order by creation timestamp")
    updated_at = Field(InfrahubOrderDirection, required=False, description="Order by updated timestamp")


class OrderInput(InputObjectType):
    disable = Boolean(required=False)
    node_metadata = Field(InfrahubNodeMetadataOrder, required=False, description="Order settings for branch metadata")


class InfrahubStandardNodeMetaData(ObjectType):
    """Base metadata type for standard nodes.

    Note: created_by and updated_by fields are added dynamically by
    GraphQLSchemaManager._patch_static_types() to use the GenericAccount interface.
    """

    created_at = DateTime(required=False, description="Date/Time the object has been created")
    updated_at = DateTime(
        required=False, description="Date/Time when the object was last modified by a user or a system task"
    )



from dataclasses import dataclass
from typing import Any

from infrahub.core.constants import NULL_VALUE, RelationshipDirection
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


@dataclass
class RelationshipFilter:
    relationship_identifier: str
    direction: RelationshipDirection

    def __hash__(self) -> int:
        return hash((self.relationship_identifier, self.direction))


@dataclass
class ProfileData:
    uuid: str
    priority: float | int
    attribute_values: dict[str, Any]
    relationship_peers: dict[RelationshipFilter, list[str]]


class GetProfileDataQuery(Query):
    type: QueryType = QueryType.READ
    insert_return: bool = False

    def __init__(
        self,
        *args: Any,
        profile_ids: list[str],
        attr_names: list[str],
        relationship_filters: list[RelationshipFilter] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.profile_ids = profile_ids
        self.attr_names = attr_names
        self.relationship_filters = relationship_filters or []

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_params)
        self.params["profile_ids"] = self.profile_ids
        self.params["attr_names"] = self.attr_names + ["profile_priority"]

        # Prepare relationship filters
        outbound_identifiers = []
        inbound_identifiers = []
        bidirectional_identifiers = []
        for rf in self.relationship_filters:
            if rf.direction == RelationshipDirection.OUTBOUND:
                outbound_identifiers.append(rf.relationship_identifier)
            elif rf.direction == RelationshipDirection.INBOUND:
                inbound_identifiers.append(rf.relationship_identifier)
            elif rf.direction == RelationshipDirection.BIDIR:
                bidirectional_identifiers.append(rf.relationship_identifier)

        self.params["outbound_identifiers"] = outbound_identifiers
        self.params["inbound_identifiers"] = inbound_identifiers
        self.params["bidirectional_identifiers"] = bidirectional_identifiers

        query = """
// --------------
// get the Profile nodes
// --------------
MATCH (profile:Node)
WHERE profile.uuid IN $profile_ids
// --------------
// make sure we only use the active ones
// --------------
CALL (profile) {
    MATCH (profile)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    RETURN r.status = "active" AS is_active
}
WITH profile
WHERE is_active = TRUE
// --------------
// get the attributes that we care about
// --------------
OPTIONAL MATCH (profile)-[:HAS_ATTRIBUTE]-(attr:Attribute)
WHERE attr.name IN $attr_names
WITH DISTINCT profile, attr
CALL (profile, attr) {
    OPTIONAL MATCH (profile)-[r:HAS_ATTRIBUTE]->(attr)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    RETURN r.status = "active" AS r1_is_active
}
// --------------
// get the attribute values
// --------------
CALL (attr) {
    OPTIONAL MATCH (attr)-[r:HAS_VALUE]->(av)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    RETURN av, r.status = "active" AS r2_is_active
    LIMIT 1
}
// --------------
// filter out null and inactive attributes
// --------------
WITH profile, CASE
    WHEN attr IS NOT NULL AND av IS NOT NULL AND r1_is_active = TRUE AND r2_is_active = TRUE THEN [attr.name, av.value]
    ELSE NULL
END AS attribute_details
WITH profile, collect(attribute_details) AS attributes
// --------------
// get all possible relationships we might want for this profile
// --------------
OPTIONAL MATCH (profile)-[r:IS_RELATED]-(rel:Relationship)
WHERE rel.name IN $outbound_identifiers + $bidirectional_identifiers + $inbound_identifiers
AND %(branch_filter)s
WITH DISTINCT profile, attributes, rel
// --------------
// filter to active near-side relationships with names and directions we want
// --------------
CALL (profile, rel) {
    OPTIONAL MATCH (profile)-[r:IS_RELATED]-(rel)
    WHERE (
        (rel.name IN $outbound_identifiers AND startNode(r) = profile)
        OR (rel.name IN $bidirectional_identifiers AND startNode(r) = profile)
        OR (rel.name IN $inbound_identifiers AND startNode(r) = rel)
    )
    AND %(branch_filter)s
    RETURN r AS r1
    ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC
    LIMIT 1
}
WITH profile, attributes, r1, rel
// --------------
// filter to active far-side relationships with names and directions we want
// --------------
CALL (profile, rel) {
    OPTIONAL MATCH (rel)-[r:IS_RELATED]-(peer)
    WHERE peer <> profile
    AND (
        (rel.name IN $outbound_identifiers AND startNode(r) = rel)
        OR (rel.name IN $bidirectional_identifiers AND startNode(r) = peer)
        OR (rel.name IN $inbound_identifiers AND startNode(r) = peer)
    )
    AND %(branch_filter)s
    RETURN r AS r2, peer
    ORDER BY r2.branch_level DESC, r2.from DESC, r2.status ASC
    LIMIT 1
}
WITH profile, attributes, r1, rel, r2, peer
// --------------
// save the direction of the relationship
// --------------
WITH *, CASE
    WHEN r1 IS NULL OR r2 IS NULL THEN NULL
    WHEN startNode(r1) = profile AND startNode(r2) = rel THEN "outbound"
    WHEN startNode(r1) = rel AND startNode(r2) = peer THEN "inbound"
    ELSE "bidirectional"
END AS direction
// --------------
// filter out null and inactive relationships
// --------------
WITH profile, attributes, CASE
    WHEN rel IS NOT NULL AND peer IS NOT NULL AND r1.status = "active" AND r2.status = "active" THEN [rel.name, direction, peer.uuid]
    ELSE NULL
END AS relationship_details
WITH profile, attributes, collect(relationship_details) AS relationships
RETURN profile.uuid AS profile_uuid, attributes, relationships
        """ % {"branch_filter": branch_filter}
        self.add_to_query(query)
        self.return_labels = ["profile_uuid", "attributes", "relationships"]

    def get_profile_data(self) -> list[ProfileData]:
        profile_data_list: list[ProfileData] = []
        for result in self.results:
            profile_uuid = result.get_as_type(label="profile_uuid", return_type=str)
            attributes = result.get(label="attributes")
            relationships = result.get(label="relationships")

            profile_data = ProfileData(
                uuid=profile_uuid, priority=float("inf"), attribute_values={}, relationship_peers={}
            )

            for attr_pair in attributes:
                if not isinstance(attr_pair, list) or len(attr_pair) != 2:
                    continue
                attr_name, attr_value = attr_pair
                if attr_value == NULL_VALUE:
                    attr_value = None
                if attr_name == "profile_priority":
                    if attr_value is not None and not isinstance(attr_value, int):
                        attr_value = int(attr_value)
                    profile_data.priority = attr_value
                else:
                    profile_data.attribute_values[attr_name] = attr_value

            # Parse relationships
            for rel_tuple in relationships:
                if not isinstance(rel_tuple, list) or len(rel_tuple) != 3:
                    continue
                rel_name, direction_str, peer_uuid = rel_tuple
                direction = RelationshipDirection(direction_str)
                rel_filter = RelationshipFilter(relationship_identifier=rel_name, direction=direction)
                if rel_filter not in profile_data.relationship_peers:
                    profile_data.relationship_peers[rel_filter] = []
                profile_data.relationship_peers[rel_filter].append(peer_uuid)

            profile_data_list.append(profile_data)
        return profile_data_list



from infrahub.core.constants import AllowOverrideType

from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema

core_file_object = GenericSchema(
    name="FileObject",
    namespace="Core",
    description="A file object for storing and managing file attachments",
    label="File Object",
    include_in_menu=False,
    attributes=[
        Attr(
            name="file_name",
            kind="Text",
            description="The name of the file as uploaded by the user",
            read_only=True,
            optional=False,
            order_weight=99990,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="checksum",
            kind="Text",
            description="SHA-1 checksum calculated on the uploaded file",
            read_only=True,
            optional=False,
            order_weight=99991,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="file_size",
            kind="Number",
            description="The size of the file in bytes",
            read_only=True,
            optional=False,
            order_weight=99992,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="file_type",
            kind="Text",
            description="The MIME type of the file",
            read_only=True,
            optional=False,
            order_weight=99993,
            allow_override=AllowOverrideType.NONE,
        ),
        Attr(
            name="storage_id",
            kind="Text",
            description="The ID of the uploaded file in Infrahub's storage",
            read_only=True,
            optional=False,
            order_weight=99994,
            allow_override=AllowOverrideType.NONE,
        ),
    ],
)



from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any

from graphql.type.definition import GraphQLNonNull
from opentelemetry import trace

from infrahub.core.constants import BranchSupportType, InfrahubKind, RelationshipHierarchyDirection
from infrahub.core.manager import NodeManager
from infrahub.core.order import OrderModel
from infrahub.exceptions import NodeNotFoundError
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.graphql.metadata import build_metadata_query_options

from ..order import deserialize_order_input
from ..parser import extract_selection
from ..permissions import get_permissions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.schema import MainSchemaTypes, NodeSchema
    from infrahub.graphql.initialization import GraphqlContext


@trace.get_tracer(__name__).start_as_current_span("account_resolver")
async def account_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
) -> dict:
    fields = extract_graphql_fields(info=info)
    graphql_context: GraphqlContext = info.context

    async with graphql_context.db.start_session(read_only=True) as db:
        results = await NodeManager.query(
            schema=InfrahubKind.GENERICACCOUNT,
            filters={"ids": [graphql_context.account_session.account_id]},
            fields=fields,
            db=db,
            order=OrderModel(disable=True),
        )
        if results:
            account_profile = await results[0].to_graphql(db=db, fields=fields)
            return account_profile

        raise NodeNotFoundError(
            node_type=InfrahubKind.GENERICACCOUNT, identifier=graphql_context.account_session.account_id
        )


@trace.get_tracer(__name__).start_as_current_span("default_resolver")
async def default_resolver(*args: Any, **kwargs) -> dict | list[dict] | None:
    """Not sure why but the default resolver returns sometime 4 positional args and sometime 2.

    When it returns 4, they are organized as follow
        - field name
        - ???
        - parent
        - info
    When it returns 2, they are organized as follow
        - parent
        - info
    """

    parent = None
    info = None
    field_name = None

    if len(args) == 4:
        parent = args[2]
        info = args[3]
        field_name = args[0]
    elif len(args) == 2:
        parent = args[0]
        info = args[1]
        field_name = info.field_name
    else:
        raise ValueError(f"expected either 2 or 4 args for default_resolver, got {len(args)}")

    # Extract the InfraHub schema by inspecting the GQL Schema
    node_schema: NodeSchema = (
        info.parent_type.of_type.graphene_type._meta.schema
        if isinstance(info.parent_type, GraphQLNonNull)
        else info.parent_type.graphene_type._meta.schema
    )

    # If the field is an attribute, return its value directly
    if field_name not in node_schema.relationship_names:
        return parent.get(field_name, None)

    # Extract the contextual information from the request context
    graphql_context: GraphqlContext = info.context

    # Extract the name of the fields in the GQL query
    fields = extract_graphql_fields(info=info)

    # Extract the schema of the node on the other end of the relationship from the GQL Schema
    node_rel = node_schema.get_relationship(info.field_name)

    # Extract only the filters from the kwargs and prepend the name of the field to the filters
    filters = {
        f"{info.field_name}__{key}": value
        for key, value in kwargs.items()
        if ("__" in key and value) or key in ["id", "ids"]
    }

    async with graphql_context.db.start_session(read_only=True) as db:
        objs = await NodeManager.query_peers(
            db=db,
            ids=[parent["id"]],
            source_kind=node_schema.kind,
            schema=node_rel,
            filters=filters,
            fields=fields,
            at=graphql_context.at,
            branch=graphql_context.branch,
            branch_agnostic=node_rel.branch is BranchSupportType.AGNOSTIC,
            fetch_peers=True,
        )

        if node_rel.cardinality == "many":
            return [
                await obj.to_graphql(db=db, fields=fields, related_node_ids=graphql_context.related_node_ids)
                for obj in objs
            ]

        # If cardinality is one
        if not objs:
            return None

        return await objs[0].to_graphql(db=db, fields=fields, related_node_ids=graphql_context.related_node_ids)


@trace.get_tracer(__name__).start_as_current_span("parent_field_name_resolver")
async def parent_field_name_resolver(parent: dict[str, dict], info: GraphQLResolveInfo) -> dict:
    """This resolver gets used when we know that the parent resolver has already gathered the required information.

    An example of this is the permissions field at the top level within default_paginated_list_resolver()
    """

    return parent[info.field_name]


def _transform_metadata_day_filters(filters: dict[str, Any]) -> dict[str, Any]:
    """Transform metadata datetime filters with 00:00:00 time into day range filters.

    When a filter like `node_metadata__created_at="2025-02-03T00:00:00"` has a time
    of exactly midnight, transform it into __after and __before filters to match
    the entire day (inclusive of midnight).

    If __after or __before filters are already explicitly defined, they will not be
    overwritten by the generated day range filters.
    """
    result = dict(filters)
    metadata_datetime_fields = ("node_metadata__created_at", "node_metadata__updated_at")

    for field in metadata_datetime_fields:
        if field not in result:
            continue
        value = result[field]
        if not isinstance(value, datetime):
            continue
        # Check if time is midnight (00:00:00)
        if value.hour == 0 and value.minute == 0 and value.second == 0 and value.microsecond == 0:
            # Remove the exact match filter
            del result[field]
            # Add __after filter with one microsecond before midnight to include objects at exactly midnight
            # Skip if __after is already explicitly defined
            after_key = f"{field}__after"
            if after_key not in result:
                one_microsecond_before = value - timedelta(microseconds=1)
                result[after_key] = one_microsecond_before
            # Add __before filter with next day (exclusive: <)
            # Skip if __before is already explicitly defined
            before_key = f"{field}__before"
            if before_key not in result:
                next_day = value + timedelta(days=1)
                result[before_key] = next_day

    return result


@trace.get_tracer(__name__).start_as_current_span("default_paginated_list_resolver")
async def default_paginated_list_resolver(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    offset: int | None = None,
    limit: int | None = None,
    order: dict | None = None,
    partial_match: bool = False,
    **kwargs: dict[str, Any],
) -> dict[str, Any]:
    schema: MainSchemaTypes = (
        info.return_type.of_type.graphene_type._meta.schema
        if isinstance(info.return_type, GraphQLNonNull)
        else info.return_type.graphene_type._meta.schema
    )

    order_model = deserialize_order_input(input_data=order)

    fields = await extract_selection(info=info, schema=schema)

    graphql_context: GraphqlContext = info.context
    async with graphql_context.db.start_session(read_only=True) as db:
        response: dict[str, Any] = {"edges": []}
        filters = {
            key: value for key, value in kwargs.items() if ("__" in key and value is not None) or key in ("ids", "hfid")
        }
        filters = _transform_metadata_day_filters(filters)

        edges: dict[str, Any] = fields.get("edges", {})
        node_fields = edges.get("node", {})
        node_metadata_fields: dict[str, Any] = edges.get("node_metadata", {})
        include_metadata = build_metadata_query_options(
            node_metadata_fields=node_metadata_fields,
            node_fields=node_fields,
        )
        if "hfid" in node_fields:
            node_fields["human_friendly_id"] = None

        permission_set: dict[str, Any] | None = None
        permissions = (
            await get_permissions(schema=schema, graphql_context=graphql_context)
            if graphql_context.permissions
            else None
        )
        if fields.get("permissions"):
            response["permissions"] = permissions

        if permissions:
            for edge in permissions["edges"]:
                if edge["node"]["kind"] == schema.kind:
                    permission_set = edge["node"]

        objs = []
        if edges or "hfid" in filters:
            objs = await NodeManager.query(
                db=db,
                schema=schema,
                filters=filters or None,
                fields=node_fields,
                include_metadata=include_metadata,
                at=graphql_context.at,
                branch=graphql_context.branch,
                limit=limit,
                offset=offset,
                partial_match=partial_match,
                order=order_model,
            )

        if "count" in fields:
            if filters.get("hfid"):
                response["count"] = len(objs)
            else:
                response["count"] = await NodeManager.count(
                    db=db,
                    schema=schema,
                    filters=filters,
                    at=graphql_context.at,
                    branch=graphql_context.branch,
                    partial_match=partial_match,
                )

        # get peer IDs for relationships
        if objs:
            objects = [
                {
                    "node": await obj.to_graphql(
                        db=db,
                        fields=node_fields,
                        related_node_ids=graphql_context.related_node_ids,
                        permissions=permission_set,
                    ),
                    "node_metadata": await obj._build_meta_response("node_metadata", edges),
                }
                for obj in objs
            ]
            response["edges"] = objects

        return response


@trace.get_tracer(__name__).start_as_current_span("single_relationship_resolver")
async def single_relationship_resolver(parent: dict, info: GraphQLResolveInfo, **kwargs: Any) -> dict[str, Any]:
    graphql_context: GraphqlContext = info.context
    resolver = graphql_context.single_relationship_resolver
    return await resolver.resolve(parent=parent, info=info, **kwargs)


@trace.get_tracer(__name__).start_as_current_span("many_relationship_resolver")
async def many_relationship_resolver(
    parent: dict, info: GraphQLResolveInfo, include_descendants: bool | None = False, **kwargs: Any
) -> dict[str, Any]:
    graphql_context: GraphqlContext = info.context
    resolver = graphql_context.many_relationship_resolver
    return await resolver.resolve(parent=parent, info=info, include_descendants=include_descendants, **kwargs)


async def ancestors_resolver(parent: dict, info: GraphQLResolveInfo, **kwargs) -> dict[str, Any]:
    return await hierarchy_resolver(
        direction=RelationshipHierarchyDirection.ANCESTORS, parent=parent, info=info, **kwargs
    )


async def descendants_resolver(parent: dict, info: GraphQLResolveInfo, **kwargs) -> dict[str, Any]:
    return await hierarchy_resolver(
        direction=RelationshipHierarchyDirection.DESCENDANTS, parent=parent, info=info, **kwargs
    )


@trace.get_tracer(__name__).start_as_current_span("hierarchy_resolver")
async def hierarchy_resolver(
    direction: RelationshipHierarchyDirection, parent: dict, info: GraphQLResolveInfo, **kwargs
) -> dict[str, Any]:
    """Resolver for ancestors and dependants for Hierarchical nodes

    This resolver is used for paginated responses and as such we redefined the requested
    fields by only reusing information below the 'node' key.
    """
    # Extract the InfraHub schema by inspecting the GQL Schema
    node_schema: NodeSchema = (
        info.parent_type.of_type.graphene_type._meta.schema
        if isinstance(info.parent_type, GraphQLNonNull)
        else info.parent_type.graphene_type._meta.schema
    )

    graphql_context: GraphqlContext = info.context

    # Extract the name of the fields in the GQL query
    fields = extract_graphql_fields(info=info)
    edges = fields.get("edges", {})
    node_fields = edges.get("node", {})

    # Extract only the filters from the kwargs and prepend the name of the field to the filters
    offset = kwargs.pop("offset", None)
    limit = kwargs.pop("limit", None)

    filters = {
        f"{info.field_name}__{key}": value
        for key, value in kwargs.items()
        if ("__" in key and value) or key in ["id", "ids"]
    }

    response: dict[str, Any] = {"edges": [], "count": None}

    async with graphql_context.db.start_session(read_only=True) as db:
        if "count" in fields:
            response["count"] = await NodeManager.count_hierarchy(
                db=db,
                id=parent["id"],
                direction=direction,
                node_schema=node_schema,
                filters=filters,
                at=graphql_context.at,
                branch=graphql_context.branch,
            )

        if not node_fields:
            return response

        objs = await NodeManager.query_hierarchy(
            db=db,
            id=parent["id"],
            direction=direction,
            node_schema=node_schema,
            filters=filters,
            fields=node_fields,
            offset=offset,
            limit=limit,
            at=graphql_context.at,
            branch=graphql_context.branch,
        )

        if not objs:
            return response
        node_graph = [await obj.to_graphql(db=db, fields=node_fields) for obj in objs.values()]

        entries = []
        for node in node_graph:
            entry = {"node": {}, "properties": {}}
            entry["node"] = node
            entries.append(entry)
        response["edges"] = entries

        return response



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.migrations.shared import GraphMigration, MigrationInput, MigrationResult
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class FixBranchLevelZeroQuery(Query):
    """Update edges with branch_level=0 to the correct branch_level.

    Edges on the global branch or default branch should have branch_level=1.
    Edges on other branches should have branch_level=2.
    """

    name = "fix_branch_level_zero"
    type: QueryType = QueryType.WRITE
    insert_return = False
    raise_error_if_empty = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["global_branch_name"] = GLOBAL_BRANCH_NAME

        query = """
        MATCH (branch_node:Branch {is_default: true})
        WITH branch_node.name AS default_branch_name
        MATCH ()-[e]->()
        WHERE e.branch_level = 0
        CALL (e, default_branch_name) {
            SET e.branch_level = CASE
                WHEN e.branch = $global_branch_name OR e.branch = default_branch_name THEN 1
                ELSE 2
            END
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class Migration053(GraphMigration):
    """
    Fix edges with branch_level=0 to have the correct branch_level.

    Edges with branch_level=0 indicate a bug where the branch level was not
    properly set during creation. This migration fixes them:
    - Edges on the global branch ("-global-") or default branch: branch_level=1
    - Edges on other branches: branch_level=2
    """

    name: str = "053_fix_branch_level_zero"
    minimum_version: int = 52
    queries: Sequence[type[Query]] = [FixBranchLevelZeroQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        return await self.do_execute(migration_input=migration_input)



from __future__ import annotations

import importlib
import json
import logging
from pathlib import Path  # noqa: TC003
from typing import TYPE_CHECKING

import typer
from graphql import parse, print_ast, print_schema
from infrahub_sdk.async_typer import AsyncTyper
from rich.logging import RichHandler

from infrahub import config
from infrahub.api.schema import SchemaLoadAPI
from infrahub.core.initialization import (
    first_time_initialization,
    initialization,
)
from infrahub.core.schema import SchemaRoot, core_models, internal_schema
from infrahub.core.schema.schema_branch import SchemaBranch
from infrahub.core.utils import delete_all_nodes
from infrahub.graphql.manager import GraphQLSchemaManager
from infrahub.graphql.schema_sort import sort_schema_ast
from infrahub.log import get_logger
from infrahub.server import app as server_app

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext

app = AsyncTyper()


@app.command(name="export-graphql-schema")
async def export_graphql_schema(
    ctx: typer.Context,  # noqa: ARG001
    config_file: str = typer.Option("infrahub.toml", envvar="INFRAHUB_CONFIG"),
    out: Path = typer.Option("schema.graphql"),  # noqa: B008
) -> None:
    """Export the Core GraphQL schema to a file."""

    config.load_and_exit(config_file_name=config_file)

    schema = SchemaRoot(**internal_schema)
    full_schema = schema.merge(schema=SchemaRoot(**core_models))

    schema_branch = SchemaBranch(cache={}, name="default")
    schema_branch.load_schema(schema=full_schema)

    schema_branch.process()

    gqlm = GraphQLSchemaManager(schema=schema_branch)
    gql_schema = gqlm.generate()

    schema_str = print_schema(gql_schema)
    schema_ast = parse(schema_str)
    sorted_schema_ast = sort_schema_ast(schema_ast)
    sorted_schema_str = print_ast(sorted_schema_ast)

    out.write_text(sorted_schema_str)


@app.command(name="export-json-schema")
async def export_json_schema(
    ctx: typer.Context,  # noqa: ARG001
    out: Path = typer.Option("openapi.json"),  # noqa: B008
) -> None:
    """Export the REST API OpenAPI schema to a file."""
    openapi_dict = server_app.openapi()
    openapi_dict["info"]["version"] = "latest"
    content = json.dumps(openapi_dict, indent=4)
    out.write_text(content)


@app.command(name="export-node-schema")
async def export_node_schema(
    ctx: typer.Context,  # noqa: ARG001
    config_file: str = typer.Option("infrahub.toml", envvar="INFRAHUB_CONFIG"),
    out: Path = typer.Option("develop.json"),  # noqa: B008
) -> None:
    """Export the repository configuration to a file."""
    config.load_and_exit(config_file_name=config_file)
    schema = SchemaLoadAPI.model_json_schema()
    schema["title"] = "InfrahubSchema"
    content = json.dumps(schema, indent=4)
    out.parent.mkdir(parents=True, exist_ok=True)
    out.write_text(content)


@app.command(name="db-init")
async def database_init(
    ctx: typer.Context,
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
) -> None:
    """Erase the content of the database and initialize it with the core schema."""

    log = get_logger()

    # --------------------------------------------------
    # CLEANUP
    #  - For now we delete everything in the database
    #   TODO, if possible try to implement this in an idempotent way
    # --------------------------------------------------

    logging.getLogger("neo4j").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    async with dbdriver.start_transaction() as db:
        log.info("Delete All Nodes")
        await delete_all_nodes(db=db)
        await first_time_initialization(db=db)

    await dbdriver.close()


@app.command(name="load-test-data")
async def load_test_data(
    ctx: typer.Context,
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
    dataset: str = "dataset01",
) -> None:
    """Load test data into the database from the `test_data` directory."""

    logging.getLogger("neo4j").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    async with dbdriver.start_session() as db:
        await initialization(db=db)

        log_level = "DEBUG"

        FORMAT = "%(message)s"
        logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
        logging.getLogger("infrahub")

        dataset_module = importlib.import_module(f"infrahub.test_data.{dataset}")
        await dataset_module.load_data(db=db)

    await dbdriver.close()



from typing import Any

from infrahub.core.diff.model.path import FrozenTrackingId
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffFreezeByProposedChangeQuery(Query):
    """Freezes DiffRoot nodes linked to a ProposedChange.

    Sets is_frozen=TRUE and updates tracking_id to 'frozen.{proposed_change_id}'
    for all DiffRoots linked via DIFF_FOR_PROPOSED_CHANGE relationship.
    """

    name = "enriched_diff_freeze_by_proposed_change"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        proposed_change_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.proposed_change_id = proposed_change_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        frozen_tracking_id = FrozenTrackingId(name=self.proposed_change_id)
        self.params = {
            "proposed_change_id": self.proposed_change_id,
            "frozen_tracking_id": frozen_tracking_id.serialize(),
        }
        query = """
MATCH (diff_root:DiffRoot)-[:DIFF_FOR_PROPOSED_CHANGE]->(pc:Node {uuid: $proposed_change_id})
SET diff_root.is_frozen = TRUE
SET diff_root.tracking_id = $frozen_tracking_id
        """
        self.add_to_query(query)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub_sdk.batch import InfrahubBatch
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.logging import get_run_logger

from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.core.constants import SYSTEM_USER_ID
from infrahub.core.migrations import MIGRATION_MAP
from infrahub.core.migrations.shared import MigrationInput
from infrahub.core.path import SchemaPath  # noqa: TC001
from infrahub.core.timestamp import Timestamp
from infrahub.workers.dependencies import get_database
from infrahub.workflows.utils import add_branch_tag

from .models import SchemaApplyMigrationData, SchemaMigrationPathResponseData

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


@flow(name="schema_apply_migrations", flow_run_name="Apply schema migrations", persist_result=True)
async def schema_apply_migrations(message: SchemaApplyMigrationData) -> list[str]:
    await add_branch_tag(branch_name=message.branch.name)
    log = get_run_logger()

    batch = InfrahubBatch()
    error_messages: list[str] = []

    if not message.migrations:
        return error_messages

    for migration in message.migrations:
        log.info(f"Preparing migration for {migration.migration_name!r} ({migration.routing_key})")

        new_node_schema: MainSchemaTypes | None = None

        if message.new_schema.has(name=migration.path.schema_kind):
            new_node_schema = message.new_schema.get(name=migration.path.schema_kind)

        if new_node_schema and new_node_schema.id:
            previous_node_schema = message.previous_schema.get_by_id(id=new_node_schema.id)
        else:
            previous_node_schema = message.previous_schema.get(name=migration.path.schema_kind)

        if not previous_node_schema:
            raise ValueError(
                f"Unable to find the previous version of the schema for {migration.path.schema_kind}, in order to run the migration."
            )

        batch.add(
            task=schema_path_migrate,
            branch=message.branch,
            migration_name=migration.migration_name,
            new_node_schema=new_node_schema,
            previous_node_schema=previous_node_schema,
            schema_path=migration.path,
            database=await get_database(),
            user_id=message.user_id,
            at=message.at,
        )

    async for _, result in batch.execute():
        error_messages.extend(result.errors)

    return error_messages


@task(  # type: ignore[arg-type]
    name="schema-path-migrate",
    task_run_name="Migrate Schema Path {migration_name} on {branch.name}",
    description="Apply a given migration to the database",
    retries=3,
    cache_policy=NONE,
)
async def schema_path_migrate(
    branch: Branch,
    migration_name: str,
    schema_path: SchemaPath,
    database: InfrahubDatabase,
    at: Timestamp,
    new_node_schema: MainSchemaTypes | None = None,
    previous_node_schema: MainSchemaTypes | None = None,
    user_id: str = SYSTEM_USER_ID,
) -> SchemaMigrationPathResponseData:
    log = get_run_logger()

    async with database.start_session() as db:
        node_kind = None
        if new_node_schema:
            node_kind = new_node_schema.kind
        elif previous_node_schema:
            node_kind = previous_node_schema.kind

        log.info(
            f"Migration for {node_kind} starting {schema_path.get_path()}",
        )
        migration_class = MIGRATION_MAP.get(migration_name)
        if not migration_class:
            raise ValueError(f"Unable to find the migration class for {migration_name}")

        migration = migration_class(  # type: ignore[call-arg]
            new_node_schema=new_node_schema,  # type: ignore[arg-type]
            previous_node_schema=previous_node_schema,  # type: ignore[arg-type]
            schema_path=schema_path,
        )
        execution_result = await migration.execute(
            migration_input=MigrationInput(db=db, at=at, user_id=user_id), branch=branch
        )

        log.info(f"Migration completed for {migration_name}")
        log.debug(f"execution_result {execution_result}")

        return SchemaMigrationPathResponseData(
            migration_name=migration_name,
            schema_path=schema_path,
            errors=execution_result.errors,
            nbr_migrations_executed=execution_result.nbr_migrations_executed,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import MainSchemaTypes

from infrahub.pools.allocator import PoolAllocator


class NoOpPoolAllocator(PoolAllocator):
    """Pool allocator that skips all allocations."""

    async def allocate_for_attribute(
        self,
        pool_relationship: RelationshipManager,  # noqa: ARG002
        target_schema: MainSchemaTypes,  # noqa: ARG002
        attribute_name: str,  # noqa: ARG002
        identifier: str,  # noqa: ARG002
    ) -> Any | None:
        return None

    async def allocate_for_relationship(self, pool_relationship: RelationshipManager, identifier: str) -> Node | None:  # noqa: ARG002
        return None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from infrahub.core.migrations.shared import (
    ArbitraryMigration,
    MigrationInput,
    MigrationResult,
    get_migration_console,
)

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

console = get_migration_console()


@dataclass
class DuplicatePoolGroup:
    node: str
    node_attribute: str
    survivor_uuid: str
    duplicate_uuids: list[str] = field(default_factory=list)


class Migration066(ArbitraryMigration):
    """Consolidate duplicate CoreNumberPool instances with pool_type='Schema'
    that share the same node + node_attribute combination.

    Keeps the earliest pool (by creation timestamp) and moves all
    IS_RESERVED and HAS_SOURCE relationships to it. Hard-deletes the duplicate pools
    Updates any NumberPoolParameters on SchemaAttributes that reference a deleted CoreNumberPool
    """

    name: str = "066_consolidate_duplicate_number_pools"
    minimum_version: int = 65

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        async with db.start_transaction() as dbt:
            try:
                duplicate_groups = await self._find_duplicate_groups(db=dbt)
                if not duplicate_groups:
                    console.log("No duplicate CoreNumberPool instances found.")
                    return MigrationResult()

                console.log(f"Found {len(duplicate_groups)} duplicate pool group(s) to consolidate.")

                pool_id_map: dict[str, str] = {}
                for group in duplicate_groups:
                    console.log(
                        f"Consolidating pools for node={group.node}, "
                        f"node_attribute={group.node_attribute}: "
                        f"keeping {group.survivor_uuid}, removing {group.duplicate_uuids}"
                    )

                    await self._reassign_is_reserved(
                        db=dbt, survivor_uuid=group.survivor_uuid, duplicate_uuids=group.duplicate_uuids
                    )
                    await self._reassign_has_source(
                        db=dbt, survivor_uuid=group.survivor_uuid, duplicate_uuids=group.duplicate_uuids
                    )
                    await self._delete_duplicate_pools(db=dbt, duplicate_uuids=group.duplicate_uuids)

                    for dup_uuid in group.duplicate_uuids:
                        pool_id_map[dup_uuid] = group.survivor_uuid

                await self._update_schema_parameters(db=dbt, pool_id_map=pool_id_map)

            except Exception as exc:
                error_msg = str(exc) or f"{type(exc).__name__}: {repr(exc)}"
                return MigrationResult(errors=[error_msg])

        return MigrationResult()

    async def _find_duplicate_groups(self, db: InfrahubDatabase) -> list[DuplicatePoolGroup]:
        """Find groups of duplicate Schema-type CoreNumberPools sharing the same node + node_attribute."""
        query = """
        MATCH (pool:CoreNumberPool)

        // -------------------------------------------
        // Get latest IS_PART_OF edge; skip pool if deleted
        // -------------------------------------------
        CALL (pool) {
            MATCH (pool)-[r_ipo:IS_PART_OF]->(root:Root)
            RETURN r_ipo
            ORDER BY r_ipo.from DESC
            LIMIT 1
        }
        WITH pool, r_ipo
        WHERE r_ipo.status = "active"

        // -------------------------------------------
        // Get latest pool_type; skip if deleted or not "Schema"
        // -------------------------------------------
        CALL (pool) {
            MATCH (pool)-[r1:HAS_ATTRIBUTE]->(attr_pt:Attribute {name: "pool_type"})-[r2:HAS_VALUE]->(av_pt)
            RETURN r1.status = "active" AND r2.status= "active" AS is_active, av_pt.value AS pool_type
            ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
            LIMIT 1
        }
        WITH pool, r_ipo, is_active, pool_type
        WHERE is_active AND pool_type = "Schema"

        // -------------------------------------------
        // Get latest node and node_attribute values
        // almost no filtering is required
        // b/c the schema is branch-agnostic and the attribute is mandatory
        // -------------------------------------------
        CALL (pool) {
            MATCH (pool)-[r1:HAS_ATTRIBUTE]->(attr_n:Attribute {name: "node"})-[r2:HAS_VALUE]->(av_n)
            RETURN av_n.value AS node_value
            ORDER BY r1.from DESC, r2.from DESC
            LIMIT 1
        }
        CALL (pool) {
            MATCH (pool)-[r1:HAS_ATTRIBUTE]->(attr_n:Attribute {name: "node_attribute"})-[r2:HAS_VALUE]->(av_na)
            RETURN av_na.value AS node_attribute_value
            ORDER BY r1.from DESC, r2.from DESC
            LIMIT 1
        }

        WITH pool.uuid AS pool_uuid,
            r_ipo.from AS created_at,
            node_value,
            node_attribute_value
        ORDER BY node_value, node_attribute_value, created_at ASC

        WITH node_value,
            node_attribute_value,
            collect(pool_uuid) AS chronological_pool_ids
        WHERE size(chronological_pool_ids) > 1

        RETURN node_value,
               node_attribute_value,
               chronological_pool_ids[0] AS survivor_uuid,
               chronological_pool_ids[1..] AS duplicate_uuids
        """

        results = await db.execute_query(query=query, name="find_duplicate_number_pools")

        return [
            DuplicatePoolGroup(
                node=record["node_value"],
                node_attribute=record["node_attribute_value"],
                survivor_uuid=record["survivor_uuid"],
                duplicate_uuids=list(record["duplicate_uuids"]),
            )
            for record in results
        ]

    async def _reassign_is_reserved(self, db: InfrahubDatabase, survivor_uuid: str, duplicate_uuids: list[str]) -> None:
        """Move IS_RESERVED relationships from duplicate pools to the surviving pool."""
        query = """
        MATCH (survivor:CoreNumberPool {uuid: $survivor_uuid})
        MATCH (dup_pool:CoreNumberPool)-[old_res:IS_RESERVED]->(target)
        WHERE dup_pool.uuid IN $duplicate_uuids

        CREATE (survivor)-[new_res:IS_RESERVED]->(target)
        SET new_res = properties(old_res)
        DELETE old_res

        RETURN count(new_res) AS moved_count
        """

        results = await db.execute_query(
            query=query,
            params={"survivor_uuid": survivor_uuid, "duplicate_uuids": duplicate_uuids},
            name="reassign_is_reserved",
        )

        if results and results[0]["moved_count"]:
            console.log(f"  Moved {results[0]['moved_count']} IS_RESERVED relationship(s) to survivor pool.")

    async def _reassign_has_source(self, db: InfrahubDatabase, survivor_uuid: str, duplicate_uuids: list[str]) -> None:
        """Move HAS_SOURCE relationships from duplicate pools to the surviving pool."""
        query = """
        MATCH (survivor:CoreNumberPool {uuid: $survivor_uuid})
        MATCH (attr:Attribute)-[old_hs:HAS_SOURCE]->(dup_pool:CoreNumberPool)
        WHERE dup_pool.uuid IN $duplicate_uuids

        CREATE (attr)-[new_hs:HAS_SOURCE]->(survivor)
        SET new_hs = properties(old_hs)
        DELETE old_hs

        RETURN count(new_hs) AS moved_count
        """

        results = await db.execute_query(
            query=query,
            params={"survivor_uuid": survivor_uuid, "duplicate_uuids": duplicate_uuids},
            name="reassign_has_source",
        )

        if results and results[0]["moved_count"]:
            console.log(f"  Moved {results[0]['moved_count']} HAS_SOURCE relationship(s) to survivor pool.")

    async def _update_schema_parameters(self, db: InfrahubDatabase, pool_id_map: dict[str, str]) -> None:
        """Update SchemaAttribute parameters JSON where number_pool_id matches a deleted pool.

        Args:
            pool_id_map: Mapping of {duplicate_uuid: survivor_uuid} for all pools to replace.
        """
        query = """
        // -------------------------------------------
        // no edge filtering required b/c we can count on the CONTAINS UUID filter
        // -------------------------------------------
        UNWIND $replacements AS replacement
        WITH replacement.dup_uuid AS dup_uuid, replacement.survivor_uuid AS survivor_uuid

        CALL (dup_uuid, survivor_uuid) {
            MATCH (sa:SchemaAttribute)-[:HAS_ATTRIBUTE]->(kind_attr:Attribute {name: "kind"})
                -[:HAS_VALUE]->(kind_val:AttributeValue {value: "NumberPool"})
            MATCH (sa)-[:HAS_ATTRIBUTE]->(param_attr:Attribute {name: "parameters"})
                -[:HAS_VALUE]->(param_val:AttributeValue)
            WHERE param_val.value CONTAINS dup_uuid

            SET param_val.value = replace(param_val.value, dup_uuid, survivor_uuid)
            RETURN 1 AS _dummy
        }

        RETURN count(*) AS updated_count
        """

        replacements = [{"dup_uuid": dup, "survivor_uuid": surv} for dup, surv in pool_id_map.items()]

        results = await db.execute_query(
            query=query,
            params={"replacements": replacements},
            name="update_schema_params_pool_id",
        )

        if results and results[0]["updated_count"]:
            console.log(f"  Updated {results[0]['updated_count']} SchemaAttribute parameter(s).")

    async def _delete_duplicate_pools(self, db: InfrahubDatabase, duplicate_uuids: list[str]) -> None:
        """Hard-delete duplicate pool nodes and their attribute sub-graphs."""
        query = """
        // ---------------------
        // only Attribute deletes are necessary b/c CoreNumberPool has no Relatonships
        // ---------------------
        MATCH (pool:CoreNumberPool)
        WHERE pool.uuid IN $duplicate_uuids
        OPTIONAL MATCH (pool)-[:HAS_ATTRIBUTE]->(attr:Attribute)

        DETACH DELETE attr, pool
        """

        await db.execute_query(
            query=query,
            params={"duplicate_uuids": duplicate_uuids},
            name="delete_duplicate_pools",
        )

        console.log(f"  Hard-deleted {len(duplicate_uuids)} duplicate pool node(s).")



from __future__ import annotations

import os
from pathlib import Path

import pytest

from .. import InfrahubClientSync
from ..utils import is_valid_url
from .loader import InfrahubYamlFile
from .utils import find_repository_config_file, load_repository_config


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("pytest-infrahub")
    group.addoption(
        "--infrahub-repo-config",
        action="store",
        dest="infrahub_repo_config",
        default=None,
        metavar="INFRAHUB_REPO_CONFIG_FILE",
        help="Infrahub configuration file for the repository (.infrahub.yml or .infrahub.yaml)",
    )
    group.addoption(
        "--infrahub-address",
        action="store",
        dest="infrahub_address",
        default="http://localhost:8000",
        metavar="INFRAHUB_TESTS_ADDRESS",
        help="Address of the Infrahub instance for live testing (default: %(default)s)",
    )
    group.addoption(
        "--infrahub-key",
        action="store",
        dest="infrahub_key",
        metavar="INFRAHUB_TESTS_API_KEY",
        default=os.getenv("INFRAHUB_API_TOKEN"),
        help="Key to use when querying the Infrahub instance for live testing",
    )
    group.addoption(
        "--infrahub-username",
        action="store",
        dest="infrahub_username",
        metavar="INFRAHUB_TESTS_USERNAME",
        help="Username to use when authenticating on the Infrahub instance for live testing",
    )
    group.addoption(
        "--infrahub-password",
        action="store",
        dest="infrahub_password",
        metavar="INFRAHUB_TESTS_PASSWORD",
        help="Password to use when authenticating on the Infrahub instance for live testing",
    )
    group.addoption(
        "--infrahub-branch",
        action="store",
        dest="infrahub_branch",
        default="main",
        metavar="INFRAHUB_TESTS_BRANCH",
        help="Branch to use when running integration tests with an Infrahub instance (default: %(default)s)",
    )


def pytest_sessionstart(session: pytest.Session) -> None:
    if session.config.option.infrahub_repo_config:
        session.infrahub_config_path = Path(session.config.option.infrahub_repo_config)  # type: ignore[attr-defined]
    else:
        session.infrahub_config_path = find_repository_config_file()  # type: ignore[attr-defined]

    if session.infrahub_config_path.is_file():  # type: ignore[attr-defined]
        session.infrahub_repo_config = load_repository_config(repo_config_file=session.infrahub_config_path)  # type: ignore[attr-defined]

    if not is_valid_url(session.config.option.infrahub_address):
        pytest.exit("Infrahub test instance address is not a valid URL", returncode=1)

    client_config = {
        "address": session.config.option.infrahub_address,
        "default_branch": session.config.option.infrahub_branch,
    }
    if hasattr(session.config.option, "infrahub_key"):
        client_config["api_token"] = session.config.option.infrahub_key
    elif hasattr(session.config.option, "infrahub_username") and hasattr(session.config.option, "infrahub_password"):
        client_config.pop("api_token")
        client_config["username"] = session.config.option.infrahub_username
        client_config["password"] = session.config.option.infrahub_password

    infrahub_client = InfrahubClientSync(config=client_config)
    session.infrahub_client = infrahub_client  # type: ignore[attr-defined]


def pytest_collect_file(parent: pytest.Collector | pytest.Item, file_path: Path) -> InfrahubYamlFile | None:
    if file_path.suffix in {".yml", ".yaml"} and file_path.name.startswith("test_"):
        return InfrahubYamlFile.from_parent(parent, path=file_path)
    return None


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line("markers", "infrahub: Infrahub test")
    config.addinivalue_line("markers", "infrahub_smoke: Smoke test for an Infrahub resource")
    config.addinivalue_line("markers", "infrahub_unit: Unit test for an Infrahub resource, works without dependencies")
    config.addinivalue_line(
        "markers",
        "infrahub_integraton: Integation test for an Infrahub resource, depends on an Infrahub running instance",
    )
    config.addinivalue_line("markers", "infrahub_check: Test related to an Infrahub Check")
    config.addinivalue_line("markers", "infrahub_graphql_query: Test related to an Infrahub GraphQL query")
    config.addinivalue_line("markers", "infrahub_jinja2_transform: Test related to an Infrahub Jinja2 Transform")
    config.addinivalue_line("markers", "infrahub_python_transform: Test related to an Infrahub Python Transform")



"""Resolvers for account metadata fields (created_by, updated_by)."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.graphql.loaders.account import AccountDataLoader, AccountLoaderParams

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch.models import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext


class AccountMetadataResolver:
    """Resolver class for account metadata fields (created_by, updated_by).

    This class maintains DataLoader instances to enable batching and caching
    of account lookups across multiple fields within the same request.
    """

    def __init__(self) -> None:
        self._data_loader_instances: dict[AccountLoaderParams, AccountDataLoader] = {}

    def _get_or_create_loader(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        fields: dict[str, Any],
    ) -> AccountDataLoader:
        """Get an existing loader or create a new one for the given parameters."""
        params = AccountLoaderParams(branch=branch, at=at, fields=fields)

        if params not in self._data_loader_instances:
            self._data_loader_instances[params] = AccountDataLoader(db=db, params=params)

        return self._data_loader_instances[params]

    async def resolve(
        self,
        parent: dict[str, Any],
        info: GraphQLResolveInfo,
    ) -> dict[str, Any] | None:
        """Resolve created_by/updated_by fields in metadata objects.

        The parent dict should contain {"id": "account-uuid", "__kind__": "CoreAccount"}
        for the field being resolved, or the field value may be None.
        """
        field_name = info.field_name  # "created_by" or "updated_by"
        account_data = parent.get(field_name)

        if not account_data or not account_data.get("id"):
            return None

        account_id = account_data["id"]
        graphql_context: GraphqlContext = info.context

        # Extract the fields requested for this account
        fields = extract_graphql_fields(info=info)

        # Get or create a loader for these parameters
        loader = self._get_or_create_loader(
            db=graphql_context.db,
            branch=graphql_context.branch,
            at=graphql_context.at,
            fields=fields,
        )

        return await loader.load(account_id)


async def account_metadata_resolver(
    parent: dict[str, Any],
    info: GraphQLResolveInfo,
) -> dict[str, Any] | None:
    """Function resolver that delegates to the AccountMetadataResolver on the context."""
    graphql_context: GraphqlContext = info.context
    resolver = graphql_context.account_metadata_resolver
    return await resolver.resolve(parent=parent, info=info)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core import registry
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.core.schema.node_schema import NodeSchema
from infrahub.log import get_logger
from infrahub.pools.schema_number_pool_upserter import SchemaNumberPoolUpserter
from infrahub.tasks.registry import update_branch_registry

from ..query import AttributeMigrationQuery, MigrationBaseQuery
from ..query.attribute_add import AttributeAddQuery
from ..shared import AttributeSchemaMigration, MigrationInput, MigrationResult

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.schema.attribute_schema import AttributeSchema
    from infrahub.database import InfrahubDatabase

    from ...branch import Branch

log = get_logger()


class NodeAttributeAddMigrationQuery01(AttributeMigrationQuery, AttributeAddQuery):
    name = "migration_node_attribute_add_01"

    def _get_node_kinds(self, schema: MainSchemaTypes, new_attribute_schema: AttributeSchema) -> list[str]:
        schema_kinds = [schema.kind]
        if not isinstance(schema, (NodeSchema, GenericSchema)):
            return schema_kinds
        if new_attribute_schema.support_profiles:
            schema_kinds.append(f"Profile{schema.kind}")
            if isinstance(schema, GenericSchema) and schema.used_by:
                schema_kinds.extend([f"Profile{kind}" for kind in schema.used_by])
        if new_attribute_schema.support_templates:
            schema_kinds.append(f"Template{schema.kind}")
            if isinstance(schema, GenericSchema) and schema.used_by:
                schema_kinds.extend([f"Template{kind}" for kind in schema.used_by])
        return schema_kinds

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ) -> None:
        node_kinds = self._get_node_kinds(
            schema=migration.new_schema, new_attribute_schema=migration.new_attribute_schema
        )
        super().__init__(
            migration=migration,
            node_kinds=node_kinds,
            attribute_name=migration.new_attribute_schema.name,
            attribute_kind=migration.new_attribute_schema.kind,
            branch_support=migration.new_attribute_schema.get_branch().value,
            default_value=migration.new_attribute_schema.default_value,
            **kwargs,
        )


class NodeAttributeAddMigration(AttributeSchemaMigration):
    name: str = "node.attribute.add"
    queries: Sequence[type[AttributeMigrationQuery]] = [NodeAttributeAddMigrationQuery01]  # type: ignore[assignment]

    async def execute(
        self,
        migration_input: MigrationInput,
        branch: Branch,
        queries: Sequence[type[MigrationBaseQuery]] | None = None,
    ) -> MigrationResult:
        if self.new_attribute_schema.inherited is True:
            return MigrationResult()
        return await super().execute(migration_input=migration_input, branch=branch, queries=queries)

    async def execute_post_queries(
        self,
        migration_input: MigrationInput,
        result: MigrationResult,
        branch: Branch,
    ) -> MigrationResult:
        if self.new_attribute_schema.kind != "NumberPool":
            return result

        db = migration_input.db
        at = migration_input.at

        upserter = SchemaNumberPoolUpserter(
            db=db,
            schema_manager=registry.schema,
        )
        number_pool = await upserter.upsert_number_pool(
            schema_node=self.new_schema,
            attribute=self.new_attribute_schema,
            branch_name=branch.name,
            at=at,
            user_id=migration_input.user_id,
        )

        await update_branch_registry(db=db, branch=branch)

        nodes: list[Node] = await registry.manager.query(
            db=db, branch=branch, schema=self.new_schema, fields={"id": True, self.new_attribute_schema.name: True}
        )

        async def allocate_numbers(db: InfrahubDatabase) -> None:
            for node in nodes:
                number = await number_pool.get_resource(  # type: ignore[attr-defined]
                    db=db, branch=branch, identifier=node.get_id(), attribute=self.new_attribute_schema, at=at
                )
                attr = node.get_attribute(name=self.new_attribute_schema.name)
                attr.value = number
                attr.set_source(number_pool.get_id())

                await node.save(db=db, fields=[self.new_attribute_schema.name], at=at)

        if db.is_transaction:
            await allocate_numbers(db=db)
        else:
            async with db.start_transaction() as dbt:
                await allocate_numbers(db=dbt)

        return result
